import hashlib
import os
import tempfile
import warnings
from zipfile import ZipFile

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from xarray import Dataset

from dtscalibration import read_apsensing_files
from dtscalibration import read_sensornet_files
from dtscalibration import read_sensortran_files
from dtscalibration import read_silixa_files
from dtscalibration.calibration.section_utils import set_sections
from dtscalibration.dts_accessor_utils import merge_double_ended
from dtscalibration.dts_accessor_utils import shift_double_ended
from dtscalibration.dts_accessor_utils import suggest_cable_shift_double_ended

np.random.seed(0)

fn = [
    "channel 1_20170921112245510.xml",
    "channel 1_20170921112746818.xml",
    "channel 1_20170921112746818.xml",
]
fn_single = [
    "channel 2_20180504132202074.xml",
    "channel 2_20180504132232903.xml",
    "channel 2_20180504132303723.xml",
]

wd = os.path.dirname(os.path.abspath(__file__))
data_dir_single_ended = os.path.join(wd, "data", "single_ended")
data_dir_double_ended = os.path.join(wd, "data", "double_ended")
data_dir_double_ended2 = os.path.join(wd, "data", "double_ended2")
data_dir_silixa_long = os.path.join(wd, "data", "double_single_ended", "channel_1")
data_dir_sensornet_single_ended = os.path.join(wd, "data", "sensornet_oryx_v3.7")
data_dir_sensornet_halo_double_ended = os.path.join(wd, "data", "sensornet_halo_v1.0")
data_dir_sensornet_oryx_double_ended = os.path.join(
    wd, "data", "sensornet_oryx_v3.7_double"
)
data_dir_sensornet_sentinel_double_ended = os.path.join(
    wd, "data", "sensornet_sentinel_v5.1_double"
)
data_dir_single_silixa_v45 = os.path.join(wd, "data", "silixa_v4.5")
data_dir_single_silixa_v7 = os.path.join(wd, "data", "silixa_v7.0")
data_dir_single_silixa_v8 = os.path.join(wd, "data", "silixa_v8.1")
data_dir_ap_sensing = os.path.join(wd, "data", "ap_sensing")
data_dir_sensortran_binary = os.path.join(wd, "data", "sensortran_binary")
data_dir_double_single_ch1 = os.path.join(
    wd, "data", "double_single_ended", "channel_1"
)
data_dir_double_single_ch2 = os.path.join(
    wd, "data", "double_single_ended", "channel_2"
)

# zips
data_dir_zipped_single_ended = os.path.join(
    wd, "data", "zipped data", "single_ended.zip"
)
data_dir_zipped_double_ended = os.path.join(
    wd, "data", "zipped data", "double_ended.zip"
)
data_dir_zipped_double_ended2 = os.path.join(
    wd, "data", "zipped data", "double_ended2.zip"
)
data_dir_zipped_silixa_long = os.path.join(
    wd, "data", "zipped data", "double_single_ended.zip"
)
data_dir_zipped_sensornet_single_ended = os.path.join(
    wd, "data", "zipped data", "sensornet_oryx_v3.7.zip"
)


def test_read_data_from_single_file_double_ended():
    """
    Check if read data from file is correct
    :return:
    """
    fp0 = os.path.join(data_dir_double_ended, fn[0])
    data = read_data_from_fp_numpy(fp0)

    nx, ncols = data.shape

    err_msg = "Not all points along the cable are read from file"
    np.testing.assert_equal(nx, 2330, err_msg=err_msg)

    err_msg = "Not all columns are read from file"
    np.testing.assert_equal(ncols, 6, err_msg=err_msg)

    actual_hash = hashlib.sha1(data).hexdigest()
    desired_hash = "51b94dedd77c83c6cdd9dd132f379a39f742edae"

    assert actual_hash == desired_hash, "The data is not read correctly"


def test_read_data_from_single_file_single_ended():
    """
    Check if read data from file is correct
    :return:
    """
    fp0 = os.path.join(data_dir_single_ended, fn_single[0])
    data = read_data_from_fp_numpy(fp0)

    nx, ncols = data.shape

    err_msg = "Not all points along the cable are read from file"
    np.testing.assert_equal(nx, 1461, err_msg=err_msg)

    err_msg = "Not all columns are read from file"
    np.testing.assert_equal(ncols, 4, err_msg=err_msg)

    actual_hash = hashlib.sha1(data).hexdigest()
    desired_hash = "58103e2d79f777f98bf279442eea138065883062"

    assert actual_hash == desired_hash, "The data is not read correctly"


def test_repr():
    ds = Dataset(
        {
            "st": (["x", "time"], np.ones((100, 5))),
            "ast": (["x", "time"], np.ones((100, 5))),
            "probe1Temperature": (["time"], range(5)),
            "probe2Temperature": (["time"], range(5)),
        },
        coords={"x": range(100), "time": range(5)},
    )

    assert "dtscalibration" in str(ds.dts)
    assert "Sections" in str(ds.dts)


def test_sections_property():
    ds = Dataset(
        {
            "st": (["x", "time"], np.ones((100, 5))),
            "ast": (["x", "time"], np.ones((100, 5))),
            "probe1Temperature": (["time"], range(5)),
            "probe2Temperature": (["time"], range(5)),
        },
        coords={"x": range(100), "time": range(5)},
    )

    sections1 = {
        "probe1Temperature": [slice(7.5, 17.0), slice(70.0, 80.0)],  # cold bath
        "probe2Temperature": [slice(24.0, 34.0), slice(85.0, 95.0)],  # warm bath
    }
    sections2 = {
        "probe1Temperature": [slice(0.0, 17.0), slice(70.0, 80.0)],  # cold bath
        "probe2Temperature": [slice(24.0, 34.0), slice(85.0, 95.0)],  # warm bath
    }
    set_sections(ds, sections1)

    assert isinstance(ds.attrs["_sections"], str)

    assert ds.dts.sections == sections1
    assert ds.dts.sections != sections2

    # test if accepts singleton numpy arrays
    set_sections(
        ds,
        {
            "probe1Temperature": [
                slice(np.array(0.0), np.array(17.0)),
                slice(70.0, 80.0),
            ]
        },
    )


def test_io_sections_property():
    ds = Dataset(
        {
            "st": (["x", "time"], np.ones((100, 5))),
            "ast": (["x", "time"], np.ones((100, 5))),
            "probe1Temperature": (["time"], range(5)),
            "probe2Temperature": (["time"], range(5)),
        },
        coords={"x": ("x", range(100), {"units": "m"}), "time": range(5)},
    )

    sections = {
        "probe1Temperature": [slice(7.5, 17.0), slice(70.0, 80.0)],  # cold bath
        "probe2Temperature": [slice(24.0, 34.0), slice(85.0, 95.0)],  # warm bath
    }
    ds["x"].attrs["units"] = "m"

    set_sections(ds, sections)

    # Create a temporary file to write data to.
    # 'with' method is used so the file is closed by tempfile
    # and free to be overwritten.
    with tempfile.NamedTemporaryFile("w") as tmp:
        temppath = tmp.name

    # Write the datastore to the temp file
    ds.to_netcdf(path=temppath)

    try:
        ds2 = xr.open_dataset(temppath)
    except ValueError as e:
        if str(e) != "cannot guess the engine, try passing one explicitly":
            raise
        warnings.warn("Could not guess engine, defaulted to netcdf4")
        ds2 = xr.open_datastore(temppath, engine="netcdf4")

    assert ds.dts.sections == ds2.dts.sections

    # Close the datastore so the temp file can be removed
    ds2.close()
    ds2 = None

    # Remove the temp file once the test is done
    if os.path.exists(temppath):
        os.remove(temppath)


def test_read_silixa_files_single_ended():
    filepath = data_dir_single_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    np.testing.assert_almost_equal(ds.st.sum(), 11387947.857, decimal=3)


def test_read_silixa_files_double_ended():
    filepath = data_dir_double_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    np.testing.assert_almost_equal(ds.st.sum(), 19613502.2617, decimal=3)


def test_read_silixa_files_single_loadinmemory():
    filepath = data_dir_single_ended

    # False
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, da.Array)

    # auto -> True
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory="auto",
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)

    # True
    ds = read_silixa_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)


def test_read_silixa_files_double_loadinmemory():
    filepath = data_dir_double_ended

    # False
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )
    for k in ["st", "ast", "rst", "rast"]:
        assert isinstance(ds[k].data, da.Array)

    # auto -> True Because small amount of data
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory="auto",
    )
    for k in ["st", "ast", "rst", "rast"]:
        assert isinstance(ds[k].data, np.ndarray)

    # True
    ds = read_silixa_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )
    for k in ["st", "ast", "rst", "rast"]:
        assert isinstance(ds[k].data, np.ndarray)


def test_read_single_silixa_v45():
    filepath = data_dir_single_silixa_v45
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, da.Array)

    ds = read_silixa_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)


def test_read_single_silixa_v7():
    filepath = data_dir_single_silixa_v7
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, da.Array)

    ds = read_silixa_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)


def test_read_single_silixa_v8():
    filepath = data_dir_single_silixa_v8
    ds = read_silixa_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, da.Array)

    ds = read_silixa_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )

    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)


@pytest.mark.skip(
    reason="Randomly fails. Has to do with delayed reading" "out of zips with dask."
)
def test_read_silixa_zipped():
    files = [
        (data_dir_zipped_single_ended, 11387947.857184),
        (data_dir_zipped_double_ended, 19613502.26171),
        (data_dir_zipped_double_ended2, 28092965.5188),
        (data_dir_zipped_silixa_long, 2.88763942e08),
    ]

    for file, stsum in files:
        with ZipFile(file) as fh:
            ds = read_silixa_files(
                zip_handle=fh,
                timezone_netcdf="UTC",
                file_ext="*.xml",
                load_in_memory=True,
            )
            np.testing.assert_almost_equal(ds.st.sum(), stsum, decimal=0)
            ds.close()


def test_read_long_silixa_files():
    filepath = data_dir_silixa_long
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")
    np.testing.assert_almost_equal(ds.st.sum(), 133223729.17096, decimal=0)


def test_read_sensornet_files_single_ended():
    filepath = data_dir_sensornet_single_ended
    ds = read_sensornet_files(
        directory=filepath,
        timezone_netcdf="UTC",
        timezone_input_files="UTC",
        add_internal_fiber_length=50.0,
        fiber_length=None,
        file_ext="*.ddf",
    )
    np.testing.assert_almost_equal(ds.st.sum(), 2955105.679, decimal=2)


def test_read_sensornet_halo_files_double_ended():
    filepath = data_dir_sensornet_halo_double_ended
    ds = read_sensornet_files(
        directory=filepath,
        timezone_netcdf="UTC",
        timezone_input_files="UTC",
        add_internal_fiber_length=50.0,
        fiber_length=1253.3,
        file_ext="*.ddf",
    )

    np.testing.assert_almost_equal(ds.st.sum(), 2835988.114, decimal=2)


def test_read_sensornet_oryx_files_double_ended():
    filepath = data_dir_sensornet_oryx_double_ended
    ds = read_sensornet_files(
        directory=filepath,
        timezone_netcdf="UTC",
        timezone_input_files="UTC",
        add_internal_fiber_length=50.0,
        fiber_length=187.0,
        file_ext="*.ddf",
    )

    np.testing.assert_almost_equal(ds.st.sum(), 2301637.154, decimal=2)
    np.testing.assert_almost_equal(ds.rst.sum(), 1835770.651, decimal=2)


def test_read_sensornet_sentinel_files_double_ended():
    filepath = data_dir_sensornet_sentinel_double_ended
    ds = read_sensornet_files(
        directory=filepath,
        timezone_netcdf="UTC",
        timezone_input_files="UTC",
        add_internal_fiber_length=50.0,
        fiber_length=2100.0,
        file_ext="*.ddf",
    )

    np.testing.assert_almost_equal(ds.st.sum(), 16531426.023, decimal=2)
    np.testing.assert_almost_equal(ds.rst.sum(), 15545880.215, decimal=2)


def test_read_apsensing_files():
    filepath = data_dir_ap_sensing
    ds = read_apsensing_files(
        directory=filepath,
        timezone_netcdf="UTC",
        timezone_input_files="UTC",
        file_ext="*.xml",
    )
    np.testing.assert_almost_equal(ds.st.sum(), 10415.2837, decimal=2)


def test_read_apsensing_files_loadinmemory():
    filepath = data_dir_ap_sensing

    # False
    ds = read_apsensing_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory=False,
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, da.Array)

    # auto -> True Because small amount of data
    ds = read_apsensing_files(
        directory=filepath,
        timezone_netcdf="UTC",
        file_ext="*.xml",
        load_in_memory="auto",
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)

    # True
    ds = read_apsensing_files(
        directory=filepath, timezone_netcdf="UTC", file_ext="*.xml", load_in_memory=True
    )
    for k in ["st", "ast"]:
        assert isinstance(ds[k].data, np.ndarray)


def test_read_sensortran_files():
    filepath = data_dir_sensortran_binary
    ds = read_sensortran_files(directory=filepath, timezone_netcdf="UTC")
    np.testing.assert_approx_equal(
        ds.st.values.astype(np.int64).sum(), np.int64(1432441254828), significant=12
    )


def read_data_from_fp_numpy(fp):
    """
    Read the data from a single Silixa xml file. Using a simple approach

    Parameters
    ----------
    fp : file, str, or pathlib.Path
        File path

    Returns
    -------
    data : ndarray
        The data of the file as numpy array of shape (nx, ncols)

    Notes
    -----
    calculating i_first and i_last is fast compared to the rest
    """

    with open(fp) as fh:
        s = fh.readlines()

    s = [si.strip() for si in s]  # remove xml hierarchy spacing

    i_first = s.index("<data>")
    i_last = len(s) - s[::-1].index("</data>") - 1

    lssl = slice(i_first + 1, i_last, 3)  # list of strings slice

    return np.loadtxt(s[lssl], delimiter=",", dtype=float)


def test_resample_datastore():
    filepath = data_dir_single_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")
    assert ds.time.size == 3

    ds_resampled = ds.resample(time="47s").mean()

    assert ds_resampled.time.size == 2

    ## No control over dim order from resample with accessor
    # assert ds_resampled.st.dims == ("x", "time"), (
    #     "The dimension have to "
    #     "be manually transposed "
    #     "after resampling. To "
    #     "guarantee the order"
    # )


def test_timeseries_keys():
    filepath = data_dir_single_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    k = ds.dts.get_timeseries_keys()

    # no false positive
    for ki in k:
        assert ds[ki].dims == ("time",)

    # no false negatives
    k_not = [ki for ki in ds.data_vars if ki not in k]
    for kni in k_not:
        assert ds[kni].dims != ("time",)


def test_shift_double_ended_shift_backforward():
    # shifting it back and forward, should result in the same
    filepath = data_dir_double_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    dsmin1 = shift_double_ended(ds, -1)
    ds2 = shift_double_ended(dsmin1, 1)

    np.testing.assert_allclose(ds.x[1:-1], ds2.x)

    for k in ds2:
        if "x" not in ds2[k].dims:
            continue

        old = ds[k].isel(x=slice(1, -1))
        new = ds2[k]

        np.testing.assert_allclose(old, new)


def test_suggest_cable_shift_double_ended():
    # need more measurements for proper testing. Therefore only checking if
    # no errors occur

    filepath = data_dir_double_ended
    ds = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    irange = np.arange(-4, 4)
    suggest_cable_shift_double_ended(ds, irange, plot_result=True)


def test_merge_double_ended():
    # Checking if alignment keeps working as designed and if the expected
    # result changed
    filepath_fw = data_dir_double_single_ch1
    filepath_bw = data_dir_double_single_ch2

    ds_fw = read_silixa_files(directory=filepath_fw)

    ds_bw = read_silixa_files(directory=filepath_bw)

    cable_length = 2017.7
    ds = merge_double_ended(ds_fw, ds_bw, cable_length=cable_length, plot_result=True)

    result = (ds.isel(time=0).st - ds.isel(time=0).rst).sum().values

    np.testing.assert_approx_equal(result, -3712866.0382, significant=10)


@pytest.mark.parametrize(
    "inotinfw, inotinbw, inotinout",
    [
        ([], [], []),
        ([1], [], [1]),
        ([], [1], [1]),
        ([1], [1], [1]),
        ([1, 2], [], [1, 2]),
        ([], [1, 2], [1, 2]),
        ([1], [2], [1, 2]),
        ([2], [1], [1, 2]),
    ],
)
def test_merge_double_ended_times(inotinfw, inotinbw, inotinout):
    """
    Arguments are the indices not included in resp fw measurements, bw measurements,
    and merged output.

    If all measurements are recorded: fw_t0, bw_t0, fw_t1, bw_t1, fw_t2, bw_t2, ..
        > all are passed

    If some are missing the accompanying measurement is skipped:
     - fw_t0, bw_t0, bw_t1, fw_t2, bw_t2, .. > fw_t0, bw_t0, fw_t2, bw_t2, ..
     - fw_t0, bw_t0, fw_t1, fw_t2, bw_t2, .. > fw_t0, bw_t0, fw_t2, bw_t2, ..
     - fw_t0, bw_t0, bw_t1, fw_t2, fw_t3, bw_t3, .. > fw_t0, bw_t0, fw_t3, bw_t3,

    Mixing forward and backward channels can be problematic when there is a pause
    after measuring all channels. This function is not perfect as the following
    situation is not caught:
     - fw_t0, bw_t0, fw_t1, bw_t2, fw_t3, bw_t3, ..
        > fw_t0, bw_t0, fw_t1, bw_t2, fw_t3, bw_t3, ..

    This routine checks that the lowest channel
    number is measured first (aka the forward channel), but it doesn't catch the
    last case as it doesn't know that fw_t1 and bw_t2 belong to different cycles.
    Any ideas are welcome.
    """
    filepath_fw = data_dir_double_single_ch1
    filepath_bw = data_dir_double_single_ch2
    cable_length = 2017.7

    ds_fw = read_silixa_files(directory=filepath_fw)
    ds_bw = read_silixa_files(directory=filepath_bw)

    # set stokes to varify proper time alignment
    ds_fw.st.values = np.tile(np.arange(ds_fw.time.size)[None], (ds_fw.x.size, 1))
    ds_bw.st.values = np.tile(np.arange(ds_bw.time.size)[None], (ds_bw.x.size, 1))

    # transform time index that is not included in: fw, bw, merged, to the ones included
    ifw = list(i for i in range(6) if i not in inotinfw)
    ibw = list(i for i in range(6) if i not in inotinbw)
    iout = list(i for i in range(6) if i not in inotinout)

    ds = merge_double_ended(
        ds_fw.isel(time=ifw),
        ds_bw.isel(time=ibw),
        cable_length=cable_length,
        plot_result=False,
        verbose=False,
    )
    assert (
        ds.time.size == len(iout) and np.all(ds.st.isel(x=0) == iout)
    ), f"FW:{ifw} & BW:{ibw} should lead to {iout} but instead leads to {ds.st.isel(x=0).values}"



import os
import warnings
from pathlib import Path
from xml.etree import ElementTree

import dask
import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from dtscalibration.io.utils import coords_time
from dtscalibration.io.utils import dim_attrs
from dtscalibration.io.utils import get_xml_namespace
from dtscalibration.io.utils import open_file
from dtscalibration.io.utils import ziphandle_to_filepathlist


def read_silixa_files(
    filepathlist=None,
    directory=None,
    zip_handle=None,
    file_ext="*.xml",
    timezone_netcdf="UTC",
    silent=False,
    load_in_memory="auto",
    **kwargs,
):
    """Read a folder with measurement files from a device of the Silixa brand. Each measurement file contains
    values for a
    single timestep. Remember to check which timezone you are working in.

    The silixa files are already timezone aware

    Parameters
    ----------
    filepathlist : list of str, optional
        List of paths that point the the silixa files
    directory : str, Path, optional
        Path to folder
    timezone_netcdf : str, optional
        Timezone string of the netcdf file. UTC follows CF-conventions.
    file_ext : str, optional
        file extension of the measurement files
    silent : bool
        If set tot True, some verbose texts are not printed to stdout/screen
    load_in_memory : {'auto', True, False}
        If 'auto' the Stokes data is only loaded to memory for small files
    kwargs : dict-like, optional
        keyword-arguments are passed to DataStore initialization

    Returns:
    --------
    datastore : DataStore
        The newly created datastore.
    """
    assert "timezone_input_files" not in kwargs, (
        "The silixa files are " "already timezone aware"
    )

    if filepathlist is None and zip_handle is None:
        filepathlist = sorted(Path(directory).glob(file_ext))

        # Make sure that the list of files contains any files
        assert len(filepathlist) >= 1, (
            "No measurement files found in provided " "directory: \n" + str(directory)
        )

    elif filepathlist is None and zip_handle:
        filepathlist = ziphandle_to_filepathlist(fh=zip_handle, extension=file_ext)

    # Make sure that the list of files contains any files
    assert len(filepathlist) >= 1, (
        "No measurement files found in provided " "list/directory"
    )

    xml_version = silixa_xml_version_check(filepathlist)

    if xml_version == 4:
        data_vars, coords, attrs = read_silixa_files_routine_v4(
            filepathlist,
            timezone_netcdf=timezone_netcdf,
            silent=silent,
            load_in_memory=load_in_memory,
        )

    elif xml_version in (6, 7, 8):
        data_vars, coords, attrs = read_silixa_files_routine_v6(
            filepathlist,
            xml_version=xml_version,
            timezone_netcdf=timezone_netcdf,
            silent=silent,
            load_in_memory=load_in_memory,
        )

    else:
        raise NotImplementedError(
            "Silixa xml version " + f"{xml_version} not implemented"
        )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="Converting non-nanosecond precision timedelta"
        )
        ds = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs, **kwargs)
    return ds


def silixa_xml_version_check(filepathlist):
    """Function which tests which version of xml files have to be read.

    Parameters
    ----------
    filepathlist

    Returns:
    --------

    """
    sep = ":"
    attrs = read_silixa_attrs_singlefile(filepathlist[0], sep)

    version_string = attrs["customData:SystemSettings:softwareVersion"]

    # Get major version from string. Tested for Ultima v4, v6, v7 XT-DTS v6
    major_version = int(version_string.replace(" ", "").split(":")[-1][0])

    return major_version


def read_silixa_attrs_singlefile(filename, sep):
    """Parameters
    ----------
    filename
    sep

    Returns:
    --------

    """
    import xmltodict

    def metakey(meta, dict_to_parse, prefix):
        """Fills the metadata dictionairy with data from dict_to_parse.
        The dict_to_parse is the raw data from a silixa xml-file.
        dict_to_parse is a nested dictionary to represent the
        different levels of hierarchy. For example,
        toplevel = {lowlevel: {key: value}}.
        This function returns {'toplevel:lowlevel:key': value}.
        Where prefix is the flattened hierarchy.

        Parameters
        ----------
        meta : dict
            the output dictionairy with prcessed metadata
        dict_to_parse : dict
        prefix

        Returns:
        --------

        """
        for key in dict_to_parse:
            if prefix == "":
                prefix_parse = key.replace("@", "")
            else:
                prefix_parse = sep.join([prefix, key.replace("@", "")])

            if prefix_parse == sep.join(("logData", "data")):
                # skip the LAF , ST data
                continue

            if hasattr(dict_to_parse[key], "keys"):
                # Nested dictionaries, flatten hierarchy.
                meta.update(metakey(meta, dict_to_parse[key], prefix_parse))

            elif isinstance(dict_to_parse[key], list):
                # if the key has values for the multiple channels
                for ival, val in enumerate(dict_to_parse[key]):
                    num_key = prefix_parse + "_" + str(ival)
                    meta.update(metakey(meta, val, num_key))
            else:
                meta[prefix_parse] = dict_to_parse[key]

        return meta

    with open_file(filename) as fh:
        doc_ = xmltodict.parse(fh.read())

    doc = doc_["wellLogs"]["wellLog"] if "wellLogs" in doc_ else doc_["logs"]["log"]

    return metakey({}, doc, "")


def read_silixa_files_routine_v6(  # noqa: MC0001
    filepathlist,
    xml_version=6,
    timezone_netcdf="UTC",
    silent=False,
    load_in_memory="auto",
):
    """Internal routine that reads Silixa files.
    Use dtscalibration.read_silixa_files function instead.

    The silixa files are already timezone aware

    Parameters
    ----------
    load_in_memory
    filepathlist
    timezone_netcdf
    silent

    Returns:
    --------

    """
    # translate names
    tld = {"ST": "st", "AST": "ast", "REV-ST": "rst", "REV-AST": "rast", "TMP": "tmp"}

    # Open the first xml file using ET, get the name space and amount of data

    with open_file(filepathlist[0]) as fh:
        xml_tree = ElementTree.parse(fh)

    namespace = get_xml_namespace(xml_tree.getroot())

    logtree = xml_tree.find(f"./{namespace}log")
    logdata_tree = logtree.find(f"./{namespace}logData")

    # Amount of datapoints is the size of the logdata tree, corrected for
    #  the mnemonic list and unit list
    nx = len(logdata_tree) - 2

    sep = ":"
    ns = {"s": namespace[1:-1]}

    # Obtain metadata from the first file
    attrs = read_silixa_attrs_singlefile(filepathlist[0], sep)

    # Add standardised required attributes
    attrs["isDoubleEnded"] = attrs["customData:isDoubleEnded"]
    double_ended_flag = bool(int(attrs["isDoubleEnded"]))

    attrs["forwardMeasurementChannel"] = attrs["customData:forwardMeasurementChannel"]
    if double_ended_flag:
        attrs["backwardMeasurementChannel"] = attrs[
            "customData:reverseMeasurementChannel"
        ]
    else:
        attrs["backwardMeasurementChannel"] = "N/A"

    # obtain basic data info
    data_item_names = (
        attrs["logData:mnemonicList"].replace(" ", "").strip(" ").split(",")
    )
    nitem = len(data_item_names)

    ntime = len(filepathlist)

    double_ended_flag = bool(int(attrs["isDoubleEnded"]))
    chFW = int(attrs["forwardMeasurementChannel"]) - 1  # zero-based
    chBW = (
        int(attrs["backwardMeasurementChannel"]) - 1 if double_ended_flag else -1
    )  # zero-based or -1 if single ended

    # print summary
    if not silent:
        print(f"{ntime} files were found, each representing a single timestep")
        print(f"{nitem} recorded vars were found: " + ", ".join(data_item_names))
        print(f"Recorded at {nx} points along the cable")

        if double_ended_flag:
            print("The measurement is double ended")
        else:
            print("The measurement is single ended")

    # obtain timeseries from data
    timeseries_loc_in_hierarchy = [
        ("log", "customData", "acquisitionTime"),
        ("log", "customData", "referenceTemperature"),
        ("log", "customData", "probe1Temperature"),
        ("log", "customData", "probe2Temperature"),
        ("log", "customData", "referenceProbeVoltage"),
        ("log", "customData", "probe1Voltage"),
        ("log", "customData", "probe2Voltage"),
        (
            "log",
            "customData",
            "UserConfiguration",
            "ChannelConfiguration",
            "AcquisitionConfiguration",
            "AcquisitionTime",
            "userAcquisitionTimeFW",
        ),
    ]

    if double_ended_flag:
        timeseries_loc_in_hierarchy.append(
            (
                "log",
                "customData",
                "UserConfiguration",
                "ChannelConfiguration",
                "AcquisitionConfiguration",
                "AcquisitionTime",
                "userAcquisitionTimeBW",
            )
        )

    timeseries = {
        item[-1]: dict(loc=item, array=np.zeros(ntime, dtype=np.float32))
        for item in timeseries_loc_in_hierarchy
    }

    # add units to timeseries (unit of measurement)
    for key, item in timeseries.items():
        item["uom"] = attrs.get(f"customData:{key}:uom", "")

    # Gather data
    arr_path = "s:" + "/s:".join(["log", "logData", "data"])

    @dask.delayed
    def grab_data_per_file(file_handle):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            eltree = ElementTree.parse(f_h)
            arr_el = eltree.findall(arr_path, namespaces=ns)

            if not len(arr_el) == nx:
                raise ValueError(
                    "Inconsistent length of x-dimension"
                    + "\nCheck if files are mixed up, or if the number of "
                    + "data points vary per file."
                )

            # remove the breaks on both sides of the string
            # split the string on the comma
            arr_str = [arr_eli.text[1:-1].split(",") for arr_eli in arr_el]

        return np.array(arr_str, dtype=float)

    data_lst_dly = [grab_data_per_file(fp) for fp in filepathlist]
    data_lst = [
        da.from_delayed(x, shape=(nx, nitem), dtype=float) for x in data_lst_dly
    ]
    data_arr = da.stack(data_lst).T  # .compute()

    # Check whether to compute data_arr (if possible 25% faster)
    data_arr_cnk = data_arr.rechunk({0: -1, 1: -1, 2: "auto"})
    if load_in_memory == "auto" and data_arr_cnk.npartitions <= 5 or load_in_memory:
        if not silent:
            print("Reading the data from disk")
        data_arr = data_arr_cnk.compute()
    else:
        if not silent:
            print("Not reading the data from disk")
        data_arr = data_arr_cnk

    data_vars = {}
    for name, data_arri in zip(data_item_names, data_arr):
        if name == "LAF":
            continue
        if tld[name] in dim_attrs:
            data_vars[tld[name]] = (["x", "time"], data_arri, dim_attrs[tld[name]])
        else:
            raise ValueError(f"Dont know what to do with the {name} data column")

    # Obtaining the timeseries data (reference temperature etc)
    _ts_dtype = [(k, np.float32) for k in timeseries]
    _time_dtype = [
        ("filename_tstamp", np.int64),
        ("minDateTimeIndex", "<U29"),
        ("maxDateTimeIndex", "<U29"),
    ]
    ts_dtype = np.dtype(_ts_dtype + _time_dtype)

    @dask.delayed
    def grab_timeseries_per_file(file_handle, xml_version):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            eltree = ElementTree.parse(f_h)

            out = []
            for _, v in timeseries.items():
                # Get all the timeseries data
                if "userAcquisitionTimeFW" in v["loc"]:
                    # requires two namespace searches
                    path1 = "s:" + "/s:".join(v["loc"][:4])
                    val1 = eltree.findall(path1, namespaces=ns)
                    path2 = "s:" + "/s:".join(v["loc"][4:6])
                    val2 = val1[chFW].find(path2, namespaces=ns)
                    out.append(val2.text)

                elif "userAcquisitionTimeBW" in v["loc"]:
                    # requires two namespace searches
                    path1 = "s:" + "/s:".join(v["loc"][:4])
                    val1 = eltree.findall(path1, namespaces=ns)
                    path2 = "s:" + "/s:".join(v["loc"][4:6])
                    val2 = val1[chBW].find(path2, namespaces=ns)
                    out.append(val2.text)

                else:
                    path = "s:" + "/s:".join(v["loc"])
                    val = eltree.find(path, namespaces=ns)
                    out.append(val.text)

            # get all the time related data
            startDateTimeIndex = eltree.find(
                "s:log/s:startDateTimeIndex", namespaces=ns
            ).text
            endDateTimeIndex = eltree.find(
                "s:log/s:endDateTimeIndex", namespaces=ns
            ).text

            if isinstance(file_handle, tuple):
                file_name = os.path.split(file_handle[0])[-1]
            else:
                file_name = os.path.split(file_handle)[-1]

            if xml_version == 6:
                tstamp = np.int64(file_name[10:27])
            elif xml_version == 7:
                tstamp = np.int64(file_name[15:27])
            elif xml_version == 8:
                tstamp = np.int64(file_name[-23:-4].replace(".", ""))
            else:
                raise ValueError(f"Unknown version number: {xml_version}")

            out += [tstamp, startDateTimeIndex, endDateTimeIndex]
        return np.array(tuple(out), dtype=ts_dtype)

    ts_lst_dly = [grab_timeseries_per_file(fp, xml_version) for fp in filepathlist]
    ts_lst = [da.from_delayed(x, shape=tuple(), dtype=ts_dtype) for x in ts_lst_dly]
    ts_arr = da.stack(ts_lst).compute()

    for name in timeseries:
        if name in dim_attrs:
            data_vars[name] = (("time",), ts_arr[name], dim_attrs[name])

        else:
            data_vars[name] = (("time",), ts_arr[name])

    # construct the coordinate dictionary
    if isinstance(filepathlist[0], tuple):
        filename_list = [os.path.split(f)[-1] for f, n2 in filepathlist]
    else:
        filename_list = [os.path.split(f)[-1] for f in filepathlist]

    coords = {
        "x": ("x", data_arr[0, :, 0], dim_attrs["x"]),
        "filename": ("time", filename_list),
        "filename_tstamp": ("time", ts_arr["filename_tstamp"]),
    }

    maxTimeIndex = pd.DatetimeIndex(ts_arr["maxDateTimeIndex"])
    dtFW = ts_arr["userAcquisitionTimeFW"].astype("timedelta64[s]")

    if not double_ended_flag:
        tcoords = coords_time(
            maxTimeIndex,
            timezone_netcdf=timezone_netcdf,
            dtFW=dtFW,
            double_ended_flag=double_ended_flag,
        )
    else:
        dtBW = ts_arr["userAcquisitionTimeBW"].astype("timedelta64[s]")
        tcoords = coords_time(
            maxTimeIndex,
            timezone_netcdf=timezone_netcdf,
            dtFW=dtFW,
            dtBW=dtBW,
            double_ended_flag=double_ended_flag,
        )

    coords.update(tcoords)

    return data_vars, coords, attrs


def read_silixa_files_routine_v4(  # noqa: MC0001
    filepathlist, timezone_netcdf="UTC", silent=False, load_in_memory="auto"
):
    """Internal routine that reads Silixa files.
    Use dtscalibration.read_silixa_files function instead.

    The silixa files are already timezone aware

    Parameters
    ----------
    load_in_memory
    filepathlist
    timezone_netcdf
    silent

    Returns:
    --------

    """
    # translate names
    tld = {"ST": "st", "AST": "ast", "REV-ST": "rst", "REV-AST": "rast", "TMP": "tmp"}

    # Open the first xml file using ET, get the name space and amount of data
    xml_tree = ElementTree.parse(filepathlist[0])
    namespace = get_xml_namespace(xml_tree.getroot())

    logtree = xml_tree.find(f"./{namespace}wellLog")
    logdata_tree = logtree.find(f"./{namespace}logData")

    # Amount of datapoints is the size of the logdata tree
    nx = len(logdata_tree)

    sep = ":"
    ns = {"s": namespace[1:-1]}

    # Obtain metadata from the first file
    attrs = read_silixa_attrs_singlefile(filepathlist[0], sep)

    # Add standardised required attributes
    attrs["isDoubleEnded"] = attrs["customData:isDoubleEnded"]
    double_ended_flag = bool(int(attrs["isDoubleEnded"]))

    attrs["forwardMeasurementChannel"] = attrs["customData:forwardMeasurementChannel"]
    if double_ended_flag:
        attrs["backwardMeasurementChannel"] = attrs[
            "customData:reverseMeasurementChannel"
        ]
    else:
        attrs["backwardMeasurementChannel"] = "N/A"

    chFW = int(attrs["forwardMeasurementChannel"]) - 1  # zero-based
    chBW = (
        int(attrs["backwardMeasurementChannel"]) - 1 if double_ended_flag else -1
    )  # zero-based or -1 if single ended

    # obtain basic data info
    if double_ended_flag:
        data_item_names = [attrs[f"logCurveInfo_{x}:mnemonic"] for x in range(6)]
    else:
        data_item_names = [attrs[f"logCurveInfo_{x}:mnemonic"] for x in range(4)]

    nitem = len(data_item_names)

    ntime = len(filepathlist)

    # print summary
    if not silent:
        print(f"{ntime} files were found, each representing a single timestep")
        print(f"{nitem} recorded vars were found: " + ", ".join(data_item_names))
        print(f"Recorded at {nx} points along the cable")

        if double_ended_flag:
            print("The measurement is double ended")
        else:
            print("The measurement is single ended")

    # obtain timeseries from data
    timeseries_loc_in_hierarchy = [
        ("wellLog", "customData", "acquisitionTime"),
        ("wellLog", "customData", "referenceTemperature"),
        ("wellLog", "customData", "probe1Temperature"),
        ("wellLog", "customData", "probe2Temperature"),
        ("wellLog", "customData", "referenceProbeVoltage"),
        ("wellLog", "customData", "probe1Voltage"),
        ("wellLog", "customData", "probe2Voltage"),
        (
            "wellLog",
            "customData",
            "UserConfiguration",
            "ChannelConfiguration",
            "AcquisitionConfiguration",
            "AcquisitionTime",
            "userAcquisitionTimeFW",
        ),
    ]

    if double_ended_flag:
        timeseries_loc_in_hierarchy.append(
            (
                "wellLog",
                "customData",
                "UserConfiguration",
                "ChannelConfiguration",
                "AcquisitionConfiguration",
                "AcquisitionTime",
                "userAcquisitionTimeBW",
            )
        )

    timeseries = {
        item[-1]: dict(loc=item, array=np.zeros(ntime, dtype=np.float32))
        for item in timeseries_loc_in_hierarchy
    }

    # add units to timeseries (unit of measurement)
    for key, item in timeseries.items():
        item["uom"] = attrs.get(f"customData:{key}:uom", "")

    # Gather data
    arr_path = "s:" + "/s:".join(["wellLog", "logData", "data"])

    @dask.delayed
    def grab_data_per_file(file_handle):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            eltree = ElementTree.parse(f_h)
            arr_el = eltree.findall(arr_path, namespaces=ns)

            if not len(arr_el) == nx:
                raise ValueError(
                    "Inconsistent length of x-dimension"
                    + "\nCheck if files are mixed up, or if the number of "
                    + "data points vary per file."
                )

            # remove the breaks on both sides of the string
            # split the string on the comma
            arr_str = [arr_eli.text.split(",") for arr_eli in arr_el]
        return np.array(arr_str, dtype=float)

    data_lst_dly = [grab_data_per_file(fp) for fp in filepathlist]
    data_lst = [
        da.from_delayed(x, shape=(nx, nitem), dtype=float) for x in data_lst_dly
    ]
    data_arr = da.stack(data_lst).T  # .compute()

    # Check whether to compute data_arr (if possible 25% faster)
    data_arr_cnk = data_arr.rechunk({0: -1, 1: -1, 2: "auto"})
    if load_in_memory == "auto" and data_arr_cnk.npartitions <= 5 or load_in_memory:
        if not silent:
            print("Reading the data from disk")
        data_arr = data_arr_cnk.compute()
    else:
        if not silent:
            print("Not reading the data from disk")
        data_arr = data_arr_cnk

    data_vars = {}
    for name, data_arri in zip(data_item_names, data_arr):
        if name == "LAF":
            continue

        if tld[name] in dim_attrs:
            data_vars[tld[name]] = (["x", "time"], data_arri, dim_attrs[tld[name]])

        else:
            raise ValueError("Dont know what to do with the" " {name} data column")

    # Obtaining the timeseries data (reference temperature etc)
    _ts_dtype = [(k, np.float32) for k in timeseries]
    _time_dtype = [
        ("filename_tstamp", np.int64),
        ("minDateTimeIndex", "<U29"),
        ("maxDateTimeIndex", "<U29"),
    ]
    ts_dtype = np.dtype(_ts_dtype + _time_dtype)

    @dask.delayed
    def grab_timeseries_per_file(file_handle):
        """Parameters
        ----------
        file_handle

        Returns:
        --------

        """
        with open_file(file_handle, mode="r") as f_h:
            eltree = ElementTree.parse(f_h)

            out = []
            for k, v in timeseries.items():
                # Get all the timeseries data
                if "userAcquisitionTimeFW" in v["loc"]:
                    # requires two namespace searches
                    path1 = "s:" + "/s:".join(v["loc"][:4])
                    val1 = eltree.findall(path1, namespaces=ns)
                    path2 = "s:" + "/s:".join(v["loc"][4:6])
                    val2 = val1[chFW].find(path2, namespaces=ns)
                    out.append(val2.text)

                elif "userAcquisitionTimeBW" in v["loc"]:
                    # requires two namespace searches
                    path1 = "s:" + "/s:".join(v["loc"][:4])
                    val1 = eltree.findall(path1, namespaces=ns)
                    path2 = "s:" + "/s:".join(v["loc"][4:6])
                    val2 = val1[chBW].find(path2, namespaces=ns)
                    out.append(val2.text)

                else:
                    path = "s:" + "/s:".join(v["loc"])
                    val = eltree.find(path, namespaces=ns)
                    out.append(val.text)

            # get all the time related data
            startDateTimeIndex = eltree.find(
                "s:wellLog/s:minDateTimeIndex", namespaces=ns
            ).text
            endDateTimeIndex = eltree.find(
                "s:wellLog/s:maxDateTimeIndex", namespaces=ns
            ).text

            if isinstance(file_handle, tuple):
                file_name = os.path.split(file_handle[0])[-1]
            else:
                file_name = os.path.split(file_handle)[-1]

            tstamp = np.int64(file_name[10:-4])

            out += [tstamp, startDateTimeIndex, endDateTimeIndex]
        return np.array(tuple(out), dtype=ts_dtype)

    ts_lst_dly = [grab_timeseries_per_file(fp) for fp in filepathlist]
    ts_lst = [da.from_delayed(x, shape=tuple(), dtype=ts_dtype) for x in ts_lst_dly]
    ts_arr = da.stack(ts_lst).compute()

    for name in timeseries:
        if name in dim_attrs:
            data_vars[name] = (("time",), ts_arr[name], dim_attrs[name])

        else:
            data_vars[name] = (("time",), ts_arr[name])

    # construct the coordinate dictionary
    coords = {
        "x": ("x", data_arr[0, :, 0], dim_attrs["x"]),
        "filename": ("time", [os.path.split(f)[1] for f in filepathlist]),
        "filename_tstamp": ("time", ts_arr["filename_tstamp"]),
    }

    maxTimeIndex = pd.DatetimeIndex(ts_arr["maxDateTimeIndex"])
    dtFW = ts_arr["userAcquisitionTimeFW"].astype("timedelta64[s]")

    if not double_ended_flag:
        tcoords = coords_time(
            maxTimeIndex,
            timezone_netcdf=timezone_netcdf,
            dtFW=dtFW,
            double_ended_flag=double_ended_flag,
        )
    else:
        dtBW = ts_arr["userAcquisitionTimeBW"].astype("timedelta64[s]")
        tcoords = coords_time(
            maxTimeIndex,
            timezone_netcdf=timezone_netcdf,
            dtFW=dtFW,
            dtBW=dtBW,
            double_ended_flag=double_ended_flag,
        )

    coords.update(tcoords)

    return data_vars, coords, attrs



import os

import numpy as np

from dtscalibration import read_silixa_files

np.random.seed(0)

fn = [
    "channel 1_20170921112245510.xml",
    "channel 1_20170921112746818.xml",
    "channel 1_20170921112746818.xml",
]
fn_single = [
    "channel 2_20180504132202074.xml",
    "channel 2_20180504132232903.xml",
    "channel 2_20180504132303723.xml",
]

if 1:
    # working dir is tests
    wd = os.path.dirname(os.path.abspath(__file__))
    data_dir_single_ended = os.path.join(wd, "data", "single_ended")
    data_dir_double_ended = os.path.join(wd, "data", "double_ended")
    data_dir_double_ended2 = os.path.join(wd, "data", "double_ended2")

else:
    # working dir is src
    data_dir_single_ended = os.path.join("..", "..", "tests", "data", "single_ended")
    data_dir_double_ended = os.path.join("..", "..", "tests", "data", "double_ended")
    data_dir_double_ended2 = os.path.join("..", "..", "tests", "data", "double_ended2")


def assert_almost_equal_verbose(actual, desired, verbose=False, **kwargs):
    """Print the actual precision decimals"""
    err = np.abs(actual - desired).max()
    dec = -np.ceil(np.log10(err))

    if not (np.isfinite(dec)):
        dec = 18.0

    m = "\n>>>>>The actual precision is: " + str(float(dec))

    if verbose:
        print(m)

    desired2 = np.broadcast_to(desired, actual.shape)
    np.testing.assert_almost_equal(actual, desired2, err_msg=m, **kwargs)
    pass


def test_average_measurements_single_ended():
    filepath = data_dir_single_ended

    ds_ = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    ds = ds_.sel(x=slice(0, 100))  # only calibrate parts of the fiber
    sections = {"probe2Temperature": [slice(6.0, 14.0)]}  # warm bath

    st_var, ast_var = 5.0, 5.0

    out = ds.dts.calibrate_single_ended(
        sections=sections, st_var=st_var, ast_var=ast_var, method="wls", solver="sparse"
    )
    ds.dts.average_monte_carlo_single_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_x_flag1=True,
        ci_avg_x_sel=slice(6.0, 14.0),
    )

    def get_section_indices(x_da, sec):
        """Returns the x-indices of the section. `sec` is a slice."""
        xis = x_da.astype(int) * 0 + np.arange(x_da.size, dtype=int)
        return xis.sel(x=sec).values

    ix = get_section_indices(ds.x, slice(6, 14))
    ds.dts.average_monte_carlo_single_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_x_flag2=True,
        ci_avg_x_isel=ix,
    )
    sl = slice(
        np.datetime64("2018-05-04T12:22:17.710000000"),
        np.datetime64("2018-05-04T12:22:47.702000000"),
    )
    ds.dts.average_monte_carlo_single_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_time_flag1=True,
        ci_avg_time_flag2=False,
        ci_avg_time_sel=sl,
    )
    ds.dts.average_monte_carlo_single_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_time_flag1=False,
        ci_avg_time_flag2=True,
        ci_avg_time_isel=range(3),
    )
    pass


def test_average_measurements_double_ended():
    filepath = data_dir_double_ended2

    ds_ = read_silixa_files(directory=filepath, timezone_netcdf="UTC", file_ext="*.xml")

    ds = ds_.sel(x=slice(0, 100))  # only calibrate parts of the fiber
    sections = {
        "probe1Temperature": [slice(7.5, 17.0), slice(70.0, 80.0)],  # cold bath
        "probe2Temperature": [slice(24.0, 34.0), slice(85.0, 95.0)],  # warm bath
    }

    st_var, ast_var, rst_var, rast_var = 5.0, 5.0, 5.0, 5.0

    out = ds.dts.calibrate_double_ended(
        sections=sections,
        st_var=st_var,
        ast_var=ast_var,
        rst_var=rst_var,
        rast_var=rast_var,
        method="wls",
        solver="sparse",
    )
    ds.dts.average_monte_carlo_double_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        rst_var=rst_var,
        rast_var=rast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_x_flag1=True,
        ci_avg_x_sel=slice(6, 10),
    )

    def get_section_indices(x_da, sec):
        """Returns the x-indices of the section. `sec` is a slice."""
        xis = x_da.astype(int) * 0 + np.arange(x_da.size, dtype=int)
        return xis.sel(x=sec).values

    ix = get_section_indices(ds.x, slice(6, 10))
    ds.dts.average_monte_carlo_double_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        rst_var=rst_var,
        rast_var=rast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_x_flag2=True,
        ci_avg_x_isel=ix,
    )
    sl = slice(
        np.datetime64("2018-03-28T00:40:54.097000000"),
        np.datetime64("2018-03-28T00:41:12.084000000"),
    )
    ds.dts.average_monte_carlo_double_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        rst_var=rst_var,
        rast_var=rast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_time_flag1=True,
        ci_avg_time_flag2=False,
        ci_avg_time_sel=sl,
    )
    ds.dts.average_monte_carlo_double_ended(
        result=out,
        st_var=st_var,
        ast_var=ast_var,
        rst_var=rst_var,
        rast_var=rast_var,
        conf_ints=[2.5, 97.5],
        mc_sample_size=50,  # <- choose a much larger sample size
        ci_avg_time_flag1=False,
        ci_avg_time_flag2=True,
        ci_avg_time_isel=range(3),
    )
    pass



import struct
from pathlib import Path
from typing import Any
from typing import Union

import numpy as np
import xarray as xr

from dtscalibration.io.utils import coords_time
from dtscalibration.io.utils import dim_attrs


def read_sensortran_files(
    directory: Union[str, Path],
    timezone_input_files: str = "UTC",
    timezone_netcdf: str = "UTC",
    silent: bool = False,
    **kwargs,
) -> xr.Dataset:
    """Read a folder with measurement files from a device of the Sensortran
    brand. Each measurement file contains values for a single timestep. Remember
    to check which timezone you are working in.

    The sensortran files are already timezone aware

    Parameters
    ----------
    directory : str, Path
        Path to folder containing BinaryRawDTS and BinaryTemp files
    timezone_input_files : str, optional
        Timezone string of the measurement files.
        Remember to check when measurements are taken.
        Also if summertime is used.
    timezone_netcdf : str, optional
        Timezone string of the netcdf file. UTC follows CF-conventions.
    silent : bool
        If set tot True, some verbose texts are not printed to stdout/screen
    kwargs : dict-like, optional
        keyword-arguments are passed to DataStore initialization

    Returns:
    --------
    DataStore
        The newly created datastore.
    """
    filepathlist_dts = sorted(Path(directory).glob("*BinaryRawDTS.dat"))

    # Make sure that the list of files contains any files
    assert len(filepathlist_dts) >= 1, (
        "No RawDTS measurement files found "
        "in provided directory: \n" + str(directory)
    )

    filepathlist_temp = [
        Path(str(f).replace("RawDTS", "Temp")) for f in filepathlist_dts
    ]

    for ii, fname in enumerate(filepathlist_dts):
        # Check if corresponding temperature file exists
        if not Path(filepathlist_temp[ii]).is_file():
            raise FileNotFoundError(
                "Could not find BinaryTemp " + f"file corresponding to {fname}"
            )

    version = sensortran_binary_version_check(filepathlist_dts)

    if version == 3:
        data_vars, coords, attrs = read_sensortran_files_routine(
            filepathlist_dts,
            filepathlist_temp,
            timezone_input_files=timezone_input_files,
            timezone_netcdf=timezone_netcdf,
            silent=silent,
        )
    else:
        raise NotImplementedError(
            "Sensortran binary version " + f"{version} not implemented"
        )

    ds = xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs, **kwargs)
    return ds


def sensortran_binary_version_check(filepathlist: list[Path]):
    """Function which tests which version the sensortran binaries are.

    Parameters
    ----------
    filepathlist

    Returns:
    --------

    """
    fname = filepathlist[0]

    with fname.open(mode="rb") as f:
        f.read(2)
        version = struct.unpack("<h", f.read(2))[0]

    return version


def read_sensortran_files_routine(
    filepathlist_dts: list[Path],
    filepathlist_temp: list[Path],
    timezone_input_files: str = "UTC",
    timezone_netcdf: str = "UTC",
    silent: bool = False,
) -> tuple[dict[str, Any], dict[str, Any], dict]:
    """Internal routine that reads sensortran files.
    Use dtscalibration.read_sensortran_files function instead.

    The sensortran files are in UTC time

    Parameters
    ----------
    filepathlist_dts
    filepathlist_temp
    timezone_netcdf
    silent

    Returns:
    --------

    """
    assert timezone_input_files == "UTC", "The sensortran files are always in UTC time."

    # Obtain metadata from the first file
    data_dts, meta_dts = read_sensortran_single(filepathlist_dts[0])
    data_temp, meta_temp = read_sensortran_single(filepathlist_temp[0])

    attrs = meta_dts

    # Add standardised required attributes
    attrs["isDoubleEnded"] = "0"

    attrs["forwardMeasurementChannel"] = meta_dts["channel_id"] - 1
    attrs["backwardMeasurementChannel"] = "N/A"

    # obtain basic data info
    nx = meta_temp["num_points"]

    ntime = len(filepathlist_dts)

    # print summary
    if not silent:
        print("%s files were found," % ntime + " each representing a single timestep")
        print("Recorded at %s points along the cable" % nx)

        print("The measurement is single ended")

    #   Gather data
    # x has already been read. should not change over time
    x = data_temp["x"]

    # Define all variables
    referenceTemperature = np.zeros(ntime)
    acquisitiontimeFW = np.ones(ntime)

    timestamp = [""] * ntime
    ST = np.zeros((nx, ntime), dtype=np.int32)
    AST = np.zeros((nx, ntime), dtype=np.int32)
    TMP = np.zeros((nx, ntime))

    ST_zero = np.zeros(ntime)
    AST_zero = np.zeros(ntime)

    for ii in range(ntime):
        data_dts, meta_dts = read_sensortran_single(filepathlist_dts[ii])
        data_temp, meta_temp = read_sensortran_single(filepathlist_temp[ii])

        timestamp[ii] = data_dts["time"]

        referenceTemperature[ii] = data_temp["reference_temperature"] - 273.15

        ST[:, ii] = data_dts["st"][:nx]
        AST[:, ii] = data_dts["ast"][:nx]
        # The TMP can vary by 1 or 2 datapoints, dynamically assign the values
        TMP[: meta_temp["num_points"], ii] = data_temp["tmp"][:nx]

        zero_index = (meta_dts["num_points"] - nx) // 2
        ST_zero[ii] = np.mean(data_dts["st"][nx + zero_index :])
        AST_zero[ii] = np.mean(data_dts["ast"][nx + zero_index :])

    data_vars = {
        "st": (["x", "time"], ST, dim_attrs["st"]),
        "ast": (["x", "time"], AST, dim_attrs["ast"]),
        "tmp": (
            ["x", "time"],
            TMP,
            {
                "name": "tmp",
                "description": "Temperature calibrated by device",
                "units": meta_temp["y_units"],
            },
        ),
        "referenceTemperature": (
            "time",
            referenceTemperature,
            {
                "name": "reference temperature",
                "description": "Internal reference " "temperature",
                "units": r"$^\circ$C",
            },
        ),
        "st_zero": (
            ["time"],
            ST_zero,
            {
                "name": "ST_zero",
                "description": "Stokes zero count",
                "units": meta_dts["y_units"],
            },
        ),
        "ast_zero": (
            ["time"],
            AST_zero,
            {
                "name": "AST_zero",
                "description": "anit-Stokes zero count",
                "units": meta_dts["y_units"],
            },
        ),
        "userAcquisitionTimeFW": (
            "time",
            acquisitiontimeFW,
            dim_attrs["userAcquisitionTimeFW"],
        ),
    }

    coords = {
        "x": (
            "x",
            x,
            {
                "name": "distance",
                "description": "Length along fiber",
                "long_description": "Starting at connector " + "of forward channel",
                "units": "m",
            },
        ),
        "filename": ("time", [f.name for f in filepathlist_dts]),
        "filename_temp": ("time", [f.name for f in filepathlist_temp]),
    }

    dtFW = data_vars["userAcquisitionTimeFW"][1].astype("timedelta64[s]")  # type: ignore

    tcoords = coords_time(
        np.array(timestamp).astype("datetime64[ns]"),
        timezone_netcdf=timezone_netcdf,
        timezone_input_files="UTC",
        dtFW=dtFW,
        double_ended_flag=False,
    )

    coords.update(tcoords)

    return data_vars, coords, attrs


def read_sensortran_single(file: Path) -> tuple[dict, dict]:
    """Internal routine that reads a single sensortran file.
    Use dtscalibration.read_sensortran_files function instead.

    Parameters
    ----------
    file

    Returns:
    --------
    data, metadata
    """
    import struct
    from datetime import datetime

    meta = {}
    data = {}
    with file.open(mode="rb") as f:
        meta["survey_type"] = struct.unpack("<h", f.read(2))[0]
        meta["hdr_version"] = struct.unpack("<h", f.read(2))[0]
        meta["x_units"] = struct.unpack("<i", f.read(4))[0]
        meta["y_units"] = struct.unpack("<i", f.read(4))[0]
        meta["num_points"] = struct.unpack("<i", f.read(4))[0]
        meta["num_pulses"] = struct.unpack("<i", f.read(4))[0]
        meta["channel_id"] = struct.unpack("<i", f.read(4))[0]
        meta["num_subtraces"] = struct.unpack("<i", f.read(4))[0]
        meta["num_skipped"] = struct.unpack("<i", f.read(4))[0]

        data["reference_temperature"] = struct.unpack("<f", f.read(4))[0]
        data["time"] = datetime.fromtimestamp(struct.unpack("<i", f.read(4))[0])

        meta["probe_name"] = f.read(128).decode("utf-16").split("\x00")[0]

        meta["hdr_size"] = struct.unpack("<i", f.read(4))[0]
        meta["hw_config"] = struct.unpack("<i", f.read(4))[0]

        data_1 = f.read(meta["num_points"] * 4)
        data_2 = f.read(meta["num_points"] * 4)

        if meta["survey_type"] == 0:
            distance = np.frombuffer(data_1, dtype=np.float32)
            temperature = np.frombuffer(data_2, dtype=np.float32)
            data["x"] = distance
            data["tmp"] = temperature

        if meta["survey_type"] == 2:
            ST = np.frombuffer(data_1, dtype=np.int32)
            AST = np.frombuffer(data_2, dtype=np.int32)
            data["st"] = ST
            data["ast"] = AST

    x_units_map = {0: "m", 1: "ft", 2: "n/a"}
    meta["x_units"] = x_units_map[meta["x_units"]]
    y_units_map = {0: "K", 1: "degC", 2: "degF", 3: "counts"}
    meta["y_units"] = y_units_map[meta["y_units"]]

    return data, meta



import numpy as np
import xarray as xr
import yaml


def set_sections(ds: xr.Dataset, sections: dict[str, list[slice]]):
    ds.attrs["_sections"] = yaml.dump(sections)


def set_matching_sections(ds: xr.Dataset, matching_sections: dict[str, list[slice]]):
    ds.attrs["_matching_sections"] = yaml.dump(matching_sections)


def validate_no_overlapping_sections(sections: dict[str, list[slice]]):
    """Check if the sections do not overlap.

    Parameters
    ----------
    sections : dict[str, list[slice]]
        The keys of the dictionary are the names of the sections.
        The values are lists of slice objects.

    Returns:
    --------
    None

    Raises:
    ------
    AssertionError
        If the sections overlap.
    """
    all_stretches = list()

    for k, v in sections.items():
        for vi in v:
            all_stretches.append(vi)

    # Check for overlapping slices
    all_start_stop = [[stretch.start, stretch.stop] for stretch in all_stretches]
    isorted_start = np.argsort([i[0] for i in all_start_stop])
    all_start_stop_startsort = [all_start_stop[i] for i in isorted_start]
    all_start_stop_startsort_flat = sum(all_start_stop_startsort, [])  # type: ignore
    assert all_start_stop_startsort_flat == sorted(
        all_start_stop_startsort_flat
    ), "Sections contains overlapping stretches"
    pass


def validate_sections_definition(sections: dict[str, list[slice]]):
    """Check if the sections are defined correctly. The sections are defined
    correctly if:
        - The keys of the sections-dictionary are strings (assertion)
        - The values of the sections-dictionary are lists (assertion).

    Parameters
    ----------
    sections : dict[str, list[slice]]
        The keys of the dictionary are the names of the sections.
        The values are lists of slice objects.

    Returns:
    --------
    None

    Raises:
    ------
    AssertionError
        If the sections are not defined correctly.
    """
    assert isinstance(sections, dict)

    for k, v in sections.items():
        assert isinstance(k, str), (
            "The keys of the " "sections-dictionary should " "be strings"
        )

        assert isinstance(v, (list, tuple)), (
            "The values of the sections-dictionary " "should be lists of slice objects."
        )


def validate_sections(ds: xr.Dataset, sections: dict[str, list[slice]]):
    """Check if the sections are valid. The sections are valid if:
        - The keys of the sections-dictionary refer to a valid timeserie
            already stored in ds.data_vars (assertion)
        - The values of the sections-dictionary are lists of slice objects.
            (assertion)
        - The slices are within the x-dimension (assertion)
        - The slices do not overlap (assertion).

    Parameters
    ----------
    ds : xr.Dataset
        The dataset that contains the timeseries that are referred to in
        the sections-dictionary.
    sections : dict[str, list[slice]]
        The keys of the dictionary are the names of the sections.
        The values are lists of slice objects.

    Returns:
    --------
    None

    Raises:
    ------
    AssertionError
        If the sections are not valid.
    """
    validate_sections_definition(sections=sections)
    validate_no_overlapping_sections(sections=sections)

    for k, v in sections.items():
        assert k in ds.data_vars, (
            "The keys of the "
            "sections-dictionary should "
            "refer to a valid timeserie "
            "already stored in "
            "ds.data_vars "
        )

        for vi in v:
            assert ds.x.sel(x=vi).size > 0, (
                f"Better define the {k} section. You tried {vi}, "
                "which is not within the x-dimension"
            )
    pass


def ufunc_per_section(
    ds: xr.Dataset,
    sections,
    func=None,
    label=None,
    subtract_from_label=None,
    temp_err=False,
    x_indices=False,
    ref_temp_broadcasted=False,
    calc_per="stretch",
    **func_kwargs,
):
    """User function applied to parts of the cable. Super useful,
    many options and slightly
    complicated.

    The function `func` is taken over all the timesteps and calculated
    per `calc_per`. This
    is returned as a dictionary

    Parameters
    ----------
    sections : Dict[str, List[slice]], optional
        If `None` is supplied, `ds.sections` is used. Define calibration
        sections. Each section requires a reference temperature time series,
        such as the temperature measured by an external temperature sensor.
        They should already be part of the DataStore object. `sections`
        is defined with a dictionary with its keywords of the
        names of the reference temperature time series. Its values are
        lists of slice objects, where each slice object is a fiber stretch
        that has the reference temperature. Afterwards, `sections` is stored
        under `ds.sections`.
    func : callable, str
        A numpy function, or lambda function to apple to each 'calc_per'.
    label
    subtract_from_label
    temp_err : bool
        The argument of the function is label minus the reference
        temperature.
    x_indices : bool
        To retreive an integer array with the indices of the
        x-coordinates in the section/stretch. The indices are sorted.
    ref_temp_broadcasted : bool
    calc_per : {'all', 'section', 'stretch'}
    func_kwargs : dict
        Dictionary with options that are passed to func

    TODO: Spend time on creating a slice instead of appendng everything\
    to a list and concatenating after.


    Returns:
    --------

    Examples:
    ---------
    1. Calculate the variance of the residuals in the along ALL the\
    reference sections wrt the temperature of the water baths

    >>> tmpf_var = d.ufunc_per_section(sections, 
    >>>     func='var',
    >>>     calc_per='all',
    >>>     label='tmpf',
    >>>     temp_err=True)

    2. Calculate the variance of the residuals in the along PER\
    reference section wrt the temperature of the water baths

    >>> tmpf_var = d.ufunc_per_section(sections, 
    >>>     func='var',
    >>>     calc_per='stretch',
    >>>     label='tmpf',
    >>>     temp_err=True)

    3. Calculate the variance of the residuals in the along PER\
    water bath wrt the temperature of the water baths

    >>> tmpf_var = d.ufunc_per_section(sections, 
    >>>     func='var',
    >>>     calc_per='section',
    >>>     label='tmpf',
    >>>     temp_err=True)

    4. Obtain the coordinates of the measurements per section

    >>> locs = d.ufunc_per_section(sections, 
    >>>     func=None,
    >>>     label='x',
    >>>     temp_err=False,
    >>>     ref_temp_broadcasted=False,
    >>>     calc_per='stretch')

    5. Number of observations per stretch

    >>> nlocs = d.ufunc_per_section(sections, 
    >>>     func=len,
    >>>     label='x',
    >>>     temp_err=False,
    >>>     ref_temp_broadcasted=False,
    >>>     calc_per='stretch')

    6. broadcast the temperature of the reference sections to\
    stretch/section/all dimensions. The value of the reference\
    temperature (a timeseries) is broadcasted to the shape of self[\
    label]. The self[label] is not used for anything else.

    >>> temp_ref = d.ufunc_per_section(sections, 
    >>>     label='st',
    >>>     ref_temp_broadcasted=True,
    >>>     calc_per='all')

    7. x-coordinate index

    >>> ix_loc = d.ufunc_per_section(sections, x_indices=True)


    Note:
    ----
    If `self[label]` or `self[subtract_from_label]` is a Dask array, a Dask
    array is returned else a numpy array is returned
    """
    from dtscalibration.dts_accessor_utils import ufunc_per_section_helper

    dataarray = None if label is None else ds[label]

    if x_indices:
        x_coords = ds.x
        reference_dataset = None
    else:
        x_coords = None
        reference_dataset = {k: ds[k] for k in sections}

    return ufunc_per_section_helper(
        x_coords=x_coords,
        sections=sections,
        func=func,
        dataarray=dataarray,
        subtract_from_dataarray=subtract_from_label,
        reference_dataset=reference_dataset,
        subtract_reference_from_dataarray=temp_err,
        ref_temp_broadcasted=ref_temp_broadcasted,
        calc_per=calc_per,
        **func_kwargs,
    )



import warnings
from typing import Optional
from typing import Union

import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import xarray as xr


def check_dims(
    ds: xr.Dataset,
    labels: Union[list[str], tuple[str]],
    correct_dims: Optional[tuple[str]] = None,
) -> None:
    """Compare the dimensions of different labels. For example: ['st', 'rst'].
    If a calculation is performed and the dimensions do not agree, the answers do not
    make sense and the matrices are broadcasted and the memory usage will explode.
    If no correct dims provided the dimensions of the different are compared.

    Parameters
    ----------
    ds :
        The DataStore object to check.
    labels :
        An iterable with labels.
    correct_dims :
        The correct dimensions.

    Returns:
    --------

    """
    if not correct_dims:
        assert len(labels) > 1, "Define the correct dimensions"

        for li in labels[1:]:
            assert ds[labels[0]].dims == ds[li].dims, (
                li + " does not have the correct dimensions."
                " Should be " + str(ds[labels[0]].dims)
            )
    else:
        for li in labels:
            assert ds[li].dims == correct_dims, (
                li + " does not have the correct dimensions. "
                "Should be " + str(correct_dims)
            )


class ParameterIndexDoubleEnded:
    """npar = 1 + 2 * nt + nx + 2 * nt * nta
    assert pval.size == npar
    assert p_var.size == npar
    if calc_cov:
        assert p_cov.shape == (npar, npar)
    gamma = pval[0]
    d_fw = pval[1:nt + 1]
    d_bw = pval[1 + nt:1 + 2 * nt]
    alpha = pval[1 + 2 * nt:1 + 2 * nt + nx]
    # store calibration parameters in DataStore
    self["gamma"] = (tuple(), gamma)
    self["alpha"] = (('x',), alpha)
    self["df"] = (('time',), d_fw)
    self["db"] = (('time',), d_bw)
    if nta > 0:
        ta = pval[1 + 2 * nt + nx:].reshape((nt, 2, nta), order='F')
        self['talpha_fw'] = (('time', 'trans_att'), ta[:, 0, :])
        self['talpha_bw'] = (('time', 'trans_att'), ta[:, 1, :]).
    """

    def __init__(self, nt, nx, nta, fix_gamma=False, fix_alpha=False):
        self.nt = nt
        self.nx = nx
        self.nta = nta
        self.fix_gamma = fix_gamma
        self.fix_alpha = fix_alpha

    @property
    def all(self):
        return np.concatenate(
            (self.gamma, self.df, self.db, self.alpha, self.ta.flatten(order="F"))
        )

    @property
    def npar(self):
        if not self.fix_gamma and not self.fix_alpha:
            return 1 + 2 * self.nt + self.nx + 2 * self.nt * self.nta
        elif self.fix_gamma and not self.fix_alpha:
            return 2 * self.nt + self.nx + 2 * self.nt * self.nta
        elif not self.fix_gamma and self.fix_alpha:
            return 1 + 2 * self.nt + 2 * self.nt * self.nta
        elif self.fix_gamma and self.fix_alpha:
            return 2 * self.nt + 2 * self.nt * self.nta

    @property
    def gamma(self):
        if self.fix_gamma:
            return []
        else:
            return [0]

    @property
    def df(self):
        if self.fix_gamma:
            return list(range(self.nt))
        else:
            return list(range(1, self.nt + 1))

    @property
    def db(self):
        if self.fix_gamma:
            return list(range(self.nt, 2 * self.nt))
        else:
            return list(range(1 + self.nt, 1 + 2 * self.nt))

    @property
    def alpha(self):
        if self.fix_alpha:
            return []
        elif self.fix_gamma:
            return list(range(2 * self.nt, 1 + 2 * self.nt + self.nx))
        elif not self.fix_gamma:
            return list(range(1 + 2 * self.nt, 1 + 2 * self.nt + self.nx))

    @property
    def ta(self):
        if self.nta == 0:
            return np.zeros((self.nt, 2, 0))
        elif not self.fix_gamma and not self.fix_alpha:
            arr = np.arange(1 + 2 * self.nt + self.nx, self.npar)
        elif self.fix_gamma and not self.fix_alpha:
            arr = np.arange(2 * self.nt + self.nx, self.npar)
        elif not self.fix_gamma and self.fix_alpha:
            arr = np.arange(1 + 2 * self.nt, self.npar)
        elif self.fix_gamma and self.fix_alpha:
            arr = np.arange(2 * self.nt, self.npar)

        return arr.reshape((self.nt, 2, self.nta), order="F")

    @property
    def taf(self):
        """Use `.reshape((nt, nta))` to convert array to (time-dim and transatt-dim). Order is the default C order.
        ta = pval[1 + 2 * nt + nx:].reshape((nt, 2, nta), order='F')
        self['talpha_fw'] = (('time', 'trans_att'), ta[:, 0, :]).
        """
        return self.ta[:, 0, :].flatten(order="C")

    @property
    def tab(self):
        """Use `.reshape((nt, nta))` to convert array to (time-dim and transatt-dim). Order is the default C order.
        ta = pval[1 + 2 * nt + nx:].reshape((nt, 2, nta), order='F')
        self['talpha_bw'] = (('time', 'trans_att'), ta[:, 1, :]).
        """
        return self.ta[:, 1, :].flatten(order="C")

    def get_ta_pars(self, pval):
        if self.nta > 0:
            if pval.ndim == 1:
                return np.take_along_axis(pval[None, None], self.ta, axis=-1)

            else:
                # assume shape is (a, npar) and returns shape (nt, 2, nta, a)
                assert pval.shape[1] == self.npar and pval.ndim == 2
                return np.stack([self.get_ta_pars(v) for v in pval], axis=-1)

        else:
            return np.zeros(shape=(self.nt, 2, 0))

    def get_taf_pars(self, pval):
        """Returns taf parameters of shape (nt, nta) or (nt, nta, a)."""
        return self.get_ta_pars(pval=pval)[:, 0, :]

    def get_tab_pars(self, pval):
        """Returns taf parameters of shape (nt, nta) or (nt, nta, a)."""
        return self.get_ta_pars(pval=pval)[:, 1, :]

    def get_taf_values(self, pval, x, trans_att, axis=""):
        """Returns taf parameters of shape (nx, nt)."""
        pval = np.atleast_2d(pval)

        assert pval.ndim == 2 and pval.shape[1] == self.npar

        arr_out = np.zeros((self.nx, self.nt))

        if self.nta == 0:
            pass

        elif axis == "":
            assert pval.shape[0] == 1

            arr = np.transpose(self.get_taf_pars(pval), axes=(1, 2, 0))  # (nta, 1, nt)

            for tai, taxi in zip(arr, trans_att):
                arr_out[x >= taxi] += tai

        elif axis == "x":
            assert pval.shape[0] == self.nx

            arr = np.transpose(self.get_taf_pars(pval), axes=(1, 2, 0))  # (nta, nx, nt)

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (x, t)
                arr_out[x >= taxi] += tai[x >= taxi]

        elif axis == "time":
            assert pval.shape[0] == self.nt

            # arr (nt, nta, nt) to have shape (nta, nt, nt)
            arr = np.transpose(self.get_taf_pars(pval), axes=(1, 2, 0))

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (t, t)
                arr_out[x >= taxi] += np.diag(tai)[None]

        return arr_out

    def get_tab_values(self, pval, x, trans_att, axis=""):
        """Returns tab parameters of shape (nx, nt)."""
        assert pval.shape[-1] == self.npar

        arr_out = np.zeros((self.nx, self.nt))

        if self.nta == 0:
            pass

        elif axis == "":
            pval = np.squeeze(pval)
            assert pval.ndim == 1
            arr = np.transpose(self.get_tab_pars(pval), axes=(1, 0))

            for tai, taxi in zip(arr, trans_att):
                arr_out[x < taxi] += tai

        elif axis == "x":
            assert pval.ndim == 2 and pval.shape[0] == self.nx

            arr = np.transpose(self.get_tab_pars(pval), axes=(1, 2, 0))

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (x, t)
                arr_out[x < taxi] += tai[x < taxi]

        elif axis == "time":
            assert pval.ndim == 2 and pval.shape[0] == self.nt

            # arr (nt, nta, nt) to have shape (nta, nt, nt)
            arr = np.transpose(self.get_tab_pars(pval), axes=(1, 2, 0))

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (t, t)
                arr_out[x < taxi] += np.diag(tai)

        return arr_out


class ParameterIndexSingleEnded:
    """if parameter fixed, they are not in
    npar = 1 + 1 + nt + nta * nt.
    """

    def __init__(self, nt, nx, nta, includes_alpha=False, includes_dalpha=True):
        assert not (
            includes_alpha and includes_dalpha
        ), "Cannot hold both dalpha and alpha"
        self.nt = nt
        self.nx = nx
        self.nta = nta
        self.includes_alpha = includes_alpha
        self.includes_dalpha = includes_dalpha

    @property
    def all(self):
        return np.concatenate(
            (self.gamma, self.dalpha, self.alpha, self.c, self.ta.flatten(order="F"))
        )

    @property
    def npar(self):
        if self.includes_alpha:
            return 1 + self.nx + self.nt + self.nta * self.nt
        elif self.includes_dalpha:
            return 1 + 1 + self.nt + self.nta * self.nt
        else:
            return 1 + self.nt + self.nta * self.nt

    @property
    def gamma(self):
        return [0]

    @property
    def dalpha(self):
        if self.includes_dalpha:
            return [1]
        else:
            return []

    @property
    def alpha(self):
        if self.includes_alpha:
            return list(range(1, 1 + self.nx))
        else:
            return []

    @property
    def c(self):
        if self.includes_alpha:
            return list(range(1 + self.nx, 1 + self.nx + self.nt))
        elif self.includes_dalpha:
            return list(range(1 + 1, 1 + 1 + self.nt))
        else:
            return list(range(1, 1 + self.nt))

    @property
    def taf(self):
        """Returns taf parameters of shape (nt, nta) or (nt, nta, a)."""
        # ta = p_val[-nt * nta:].reshape((nt, nta), order='F')
        # self["talpha"] = (('time', 'trans_att'), ta[:, :])
        if self.includes_alpha:
            return np.arange(
                1 + self.nx + self.nt, 1 + self.nx + self.nt + self.nt * self.nta
            ).reshape((self.nt, self.nta), order="F")
        elif self.includes_dalpha:
            return np.arange(
                1 + 1 + self.nt, 1 + 1 + self.nt + self.nt * self.nta
            ).reshape((self.nt, self.nta), order="F")
        else:
            return np.arange(1 + self.nt, 1 + self.nt + self.nt * self.nta).reshape(
                (self.nt, self.nta), order="F"
            )

    def get_taf_pars(self, pval):
        if self.nta > 0:
            if pval.ndim == 1:
                # returns shape (nta, nt)
                assert len(pval) == self.npar, "Length of pval is incorrect"
                return np.stack([pval[tafi] for tafi in self.taf.T])

            else:
                # assume shape is (a, npar) and returns shape (nta, nt, a)
                assert pval.shape[1] == self.npar and pval.ndim == 2
                return np.stack([self.get_taf_pars(v) for v in pval], axis=-1)

        else:
            return np.zeros(shape=(self.nt, 0))

    def get_taf_values(self, pval, x, trans_att, axis=""):
        """Returns taf parameters of shape (nx, nt)."""
        # assert pval.ndim == 2 and pval.shape[1] == self.npar

        arr_out = np.zeros((self.nx, self.nt))

        if self.nta == 0:
            pass

        elif axis == "":
            pval = pval.flatten()
            assert pval.shape == (self.npar,)
            arr = self.get_taf_pars(pval)
            assert arr.shape == (
                self.nta,
                self.nt,
            )

            for tai, taxi in zip(arr, trans_att):
                arr_out[x >= taxi] += tai

        elif axis == "x":
            assert pval.shape == (self.nx, self.npar)
            arr = self.get_taf_pars(pval)
            assert arr.shape == (self.nta, self.nx, self.nt)

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (x, t)
                arr_out[x >= taxi] += tai[x >= taxi]

        elif axis == "time":
            assert pval.shape == (self.nt, self.npar)
            arr = self.get_taf_pars(pval)
            assert arr.shape == (self.nta, self.nt, self.nt)

            for tai, taxi in zip(arr, trans_att):
                # loop over nta, tai has shape (t, t)
                arr_out[x >= taxi] += np.diag(tai)[None]

        return arr_out


def get_netcdf_encoding(
    ds: xr.Dataset, zlib: bool = True, complevel: int = 5, **kwargs
) -> dict:
    """Get default netcdf compression parameters. The same for each data variable.

    TODO: Truncate precision to XML precision per data variable


    Parameters
    ----------
    zlib
    complevel
    ds : DataStore

    Returns:
    --------
    encoding:
        Encoding dictionary.
    """
    comp = dict(zlib=zlib, complevel=complevel)
    comp.update(kwargs)
    encoding = {var: comp for var in ds.data_vars}

    return encoding


def get_params_from_pval_double_ended(ip, coords, p_val=None, p_cov=None):
    if p_val is not None:
        assert len(p_val) == ip.npar, "Length of p_val is incorrect"

        params = xr.Dataset(coords=coords)

        # save estimates and variances to datastore, skip covariances
        params["gamma"] = (tuple(), p_val[ip.gamma].item())
        params["alpha"] = (("x",), p_val[ip.alpha])
        params["df"] = (("time",), p_val[ip.df])
        params["db"] = (("time",), p_val[ip.db])

        if ip.nta:
            params["talpha_fw"] = (
                ("time", "trans_att"),
                p_val[ip.taf].reshape((ip.nt, ip.nta), order="C"),
            )
            params["talpha_bw"] = (
                ("time", "trans_att"),
                p_val[ip.tab].reshape((ip.nt, ip.nta), order="C"),
            )
        else:
            params["talpha_fw"] = (("time", "trans_att"), np.zeros((ip.nt, 0)))
            params["talpha_bw"] = (("time", "trans_att"), np.zeros((ip.nt, 0)))

        params["talpha_fw_full"] = (
            ("x", "time"),
            ip.get_taf_values(
                pval=p_val,
                x=params.x.values,
                trans_att=params.trans_att.values,
                axis="",
            ),
        )
        params["talpha_bw_full"] = (
            ("x", "time"),
            ip.get_tab_values(
                pval=p_val,
                x=params.x.values,
                trans_att=params.trans_att.values,
                axis="",
            ),
        )
    if p_cov is not None:
        assert p_cov.shape == (ip.npar, ip.npar), "Shape of p_cov is incorrect"

        # extract covariances and ensure broadcastable to (nx, nt)
        params["gamma_df"] = (("time",), p_cov[np.ix_(ip.gamma, ip.df)][0])
        params["gamma_db"] = (("time",), p_cov[np.ix_(ip.gamma, ip.db)][0])
        params["gamma_alpha"] = (("x",), p_cov[np.ix_(ip.alpha, ip.gamma)][:, 0])
        params["df_db"] = (
            ("time",),
            p_cov[ip.df, ip.db],
        )
        params["alpha_df"] = (
            (
                "x",
                "time",
            ),
            p_cov[np.ix_(ip.alpha, ip.df)],
        )
        params["alpha_db"] = (
            (
                "x",
                "time",
            ),
            p_cov[np.ix_(ip.alpha, ip.db)],
        )
        params["tafw_gamma"] = (
            (
                "x",
                "time",
            ),
            ip.get_taf_values(
                pval=p_cov[ip.gamma],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="",
            ),
        )
        params["tabw_gamma"] = (
            (
                "x",
                "time",
            ),
            ip.get_tab_values(
                pval=p_cov[ip.gamma],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="",
            ),
        )
        params["tafw_alpha"] = (
            (
                "x",
                "time",
            ),
            ip.get_taf_values(
                pval=p_cov[ip.alpha],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="x",
            ),
        )
        params["tabw_alpha"] = (
            (
                "x",
                "time",
            ),
            ip.get_tab_values(
                pval=p_cov[ip.alpha],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="x",
            ),
        )
        params["tafw_df"] = (
            (
                "x",
                "time",
            ),
            ip.get_taf_values(
                pval=p_cov[ip.df],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="time",
            ),
        )
        params["tafw_db"] = (
            (
                "x",
                "time",
            ),
            ip.get_taf_values(
                pval=p_cov[ip.db],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="time",
            ),
        )
        params["tabw_db"] = (
            (
                "x",
                "time",
            ),
            ip.get_tab_values(
                pval=p_cov[ip.db],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="time",
            ),
        )
        params["tabw_df"] = (
            (
                "x",
                "time",
            ),
            ip.get_tab_values(
                pval=p_cov[ip.df],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="time",
            ),
        )
        # sigma2_tafw_tabw
    return params


def get_params_from_pval_single_ended(
    ip, coords, p_val=None, p_var=None, p_cov=None, fix_alpha=None
):
    if p_val is not None:
        assert len(p_val) == ip.npar, "Length of p_val is incorrect"

    if p_var is not None:
        assert len(p_var) == ip.npar, "Length of p_var is incorrect"

    if p_cov is not None:
        assert p_cov.shape == (ip.npar, ip.npar), "Shape of p_cov is incorrect"

    params = xr.Dataset(coords=coords)
    param_covs = xr.Dataset(coords=coords)

    params["gamma"] = (tuple(), p_val[ip.gamma].item())
    param_covs["gamma"] = (tuple(), p_var[ip.gamma].item())

    if ip.nta > 0:
        params["talpha_fw"] = (
            ("trans_att", "time"),
            ip.get_taf_pars(p_val),
        )
        param_covs["talpha_fw"] = (
            ("trans_att", "time"),
            ip.get_taf_pars(p_var),
        )
    else:
        params["talpha_fw"] = (("trans_att", "time"), np.zeros((0, ip.nt)))
        param_covs["talpha_fw"] = (("trans_att", "time"), np.zeros((0, ip.nt)))

    params["c"] = (("time",), p_val[ip.c])
    param_covs["c"] = (("time",), p_var[ip.c])

    if fix_alpha is not None:
        params["alpha"] = (("x",), fix_alpha[0])
        param_covs["alpha"] = (("x",), fix_alpha[1])

    else:
        params["dalpha"] = (tuple(), p_val[ip.dalpha].item())
        param_covs["dalpha"] = (tuple(), p_var[ip.dalpha].item())

        params["alpha"] = params["dalpha"] * params["x"]
        param_covs["alpha"] = param_covs["dalpha"] * params["x"] ** 2

        param_covs["gamma_dalpha"] = (tuple(), p_cov[np.ix_(ip.dalpha, ip.gamma)][0, 0])
        param_covs["dalpha_c"] = (("time",), p_cov[np.ix_(ip.dalpha, ip.c)][0, :])
        param_covs["tafw_dalpha"] = (
            ("x", "time"),
            ip.get_taf_values(
                pval=p_cov[ip.dalpha],
                x=params["x"].values,
                trans_att=params["trans_att"].values,
                axis="",
            ),
        )

    params["talpha_fw_full"] = (
        ("x", "time"),
        ip.get_taf_values(
            pval=p_val,
            x=params["x"].values,
            trans_att=params["trans_att"].values,
            axis="",
        ),
    )
    param_covs["talpha_fw_full"] = (
        ("x", "time"),
        ip.get_taf_values(
            pval=p_var,
            x=params["x"].values,
            trans_att=params["trans_att"].values,
            axis="",
        ),
    )
    param_covs["gamma_c"] = (("time",), p_cov[np.ix_(ip.gamma, ip.c)][0, :])
    param_covs["tafw_gamma"] = (
        ("x", "time"),
        ip.get_taf_values(
            pval=p_cov[ip.gamma],
            x=params["x"].values,
            trans_att=params["trans_att"].values,
            axis="",
        ),
    )
    param_covs["tafw_c"] = (
        ("x", "time"),
        ip.get_taf_values(
            pval=p_cov[ip.c],
            x=params["x"].values,
            trans_att=params["trans_att"].values,
            axis="time",
        ),
    )
    return params, param_covs


def merge_double_ended(
    ds_fw: xr.Dataset,
    ds_bw: xr.Dataset,
    cable_length: float,
    plot_result: bool = True,
    verbose: bool = True,
) -> xr.Dataset:
    """Some measurements are not set up on the DTS-device as double-ended
    meausurements. This means that the two channels have to be merged manually.

    This function can merge two single-ended DataStore objects into a single
    double-ended DataStore. There is no interpolation, the two arrays are
    flipped and overlayed, based on the entered cable length. This can
    introduce spatial inaccuracies with extremely long cables.

    Parameters
    ----------
    ds_fw : DataSore object
        DataStore object representing the forward measurement channel
    ds_bw : DataSore object
        DataStore object representing the backward measurement channel
    cable_length : float
        Manually estimated cable length to base alignment on
    plot_result : bool
        Plot the aligned Stokes of the forward and backward channels
    verbose : bool

    Returns:
    --------
    ds : DataStore object
        With the two channels merged
    """
    assert (
        ds_fw.attrs["isDoubleEnded"] == "0" and ds_bw.attrs["isDoubleEnded"] == "0"
    ), "(one of the) input DataStores is already double ended"

    ds_fw, ds_bw = merge_double_ended_times(ds_fw, ds_bw, verbose=verbose)

    ds = ds_fw.copy()
    ds_bw = ds_bw.copy()

    ds_bw["x"] = cable_length - ds_bw["x"].values

    # TODO: check if reindexing matters, and should be used.
    # one way to do it is performed below, but this could create artifacts
    x_resolution = ds["x"].values[1] - ds["x"].values[0]
    ds_bw = ds_bw.reindex(
        {"x": ds["x"]}, method="nearest", tolerance=0.99 * x_resolution
    )

    ds_bw = ds_bw.sortby("x")

    ds["rst"] = (["x", "time"], ds_bw["st"].values)
    ds["rast"] = (["x", "time"], ds_bw["ast"].values)

    ds = ds.dropna(dim="x")

    ds.attrs["isDoubleEnded"] = "1"
    ds["userAcquisitionTimeBW"] = ("time", ds_bw["userAcquisitionTimeFW"].values)

    if plot_result:
        _, ax = plt.subplots()
        ds["st"].isel(time=0).plot(ax=ax, label="Stokes forward")
        ds["rst"].isel(time=0).plot(ax=ax, label="Stokes backward")
        ax.legend()

    return ds


def merge_double_ended_times(
    ds_fw: xr.Dataset,
    ds_bw: xr.Dataset,
    verify_timedeltas: bool = True,
    verbose: bool = True,
) -> tuple[xr.Dataset, xr.Dataset]:
    """Helper for `merge_double_ended()` to deal with missing measurements. The
    number of measurements of the forward and backward channels might get out
    of sync if the device shuts down before the measurement of the last channel
    is complete. This skips all measurements that are not accompanied by a partner
    channel.

    Provides little protection against swapping fw and bw.

    If all measurements are recorded: fw_t0, bw_t0, fw_t1, bw_t1, fw_t2, bw_t2, ..
        > all are passed

    If some are missing the accompanying measurement is skipped:
     - fw_t0, bw_t0, bw_t1, fw_t2, bw_t2, .. > fw_t0, bw_t0, fw_t2, bw_t2, ..
     - fw_t0, bw_t0, fw_t1, fw_t2, bw_t2, .. > fw_t0, bw_t0, fw_t2, bw_t2, ..
     - fw_t0, bw_t0, bw_t1, fw_t2, fw_t3, bw_t3, .. > fw_t0, bw_t0, fw_t3, bw_t3,

    Mixing forward and backward channels can be problematic when there is a pause
    after measuring all channels. This function is not perfect as the following
    situation is not caught:
     - fw_t0, bw_t0, fw_t1, bw_t2, fw_t3, bw_t3, ..
        > fw_t0, bw_t0, fw_t1, bw_t2, fw_t3, bw_t3, ..

    This routine checks that the lowest channel
    number is measured first (aka the forward channel), but it doesn't catch the
    last case as it doesn't know that fw_t1 and bw_t2 belong to different cycles.

    Parameters
    ----------
    ds_fw : DataSore object
        DataStore object representing the forward measurement channel
    ds_bw : DataSore object
        DataStore object representing the backward measurement channel
    verify_timedeltas : bool
        Check whether times between forward and backward measurements are similar to
        those of neighboring measurements
    verbose : bool
        Print additional information to screen

    Returns:
    --------
    ds_fw_sel : DataSore object
        DataStore object representing the forward measurement channel with
        only times for which there is also a ds_bw measurement
    ds_bw_sel : DataSore object
        DataStore object representing the backward measurement channel with
        only times for which there is also a ds_fw measurement
    """
    if "forward channel" in ds_fw.attrs and "forward channel" in ds_bw.attrs:
        assert (
            ds_fw.attrs["forward channel"] < ds_bw.attrs["forward channel"]
        ), "ds_fw and ds_bw are swapped"
    elif (
        "forwardMeasurementChannel" in ds_fw.attrs
        and "forwardMeasurementChannel" in ds_bw.attrs
    ):
        assert (
            ds_fw.attrs["forwardMeasurementChannel"]
            < ds_bw.attrs["forwardMeasurementChannel"]
        ), "ds_fw and ds_bw are swapped"

    # Are all dt's within 1.5 seconds from one another?
    if (ds_bw.time.size == ds_fw.time.size) and np.all(
        ds_bw.time.values > ds_fw.time.values
    ):
        if verify_timedeltas:
            dt_ori = (ds_bw.time.values - ds_fw.time.values) / np.array(
                1, dtype="timedelta64[s]"
            )
            dt_all_close = np.allclose(dt_ori, dt_ori[0], atol=1.5, rtol=0.0)
        else:
            dt_all_close = True

        if dt_all_close:
            return ds_fw, ds_bw

    iuse_chfw = list()
    iuse_chbw = list()

    times_fw = {k: ("fw", i) for i, k in enumerate(ds_fw.time.values)}
    times_bw = {k: ("bw", i) for i, k in enumerate(ds_bw.time.values)}
    times_all = dict(sorted(({**times_fw, **times_bw}).items()))
    times_all_val = list(times_all.values())

    for (direction, ind), (direction_next, ind_next) in zip(
        times_all_val[:-1], times_all_val[1:]
    ):
        if direction == "fw" and direction_next == "bw":
            iuse_chfw.append(ind)
            iuse_chbw.append(ind_next)

        elif direction == "bw" and direction_next == "fw":
            pass

        elif direction == "fw" and direction_next == "fw" and verbose:
            print(
                f"Missing backward measurement beween {ds_fw.time.values[ind]} and {ds_fw.time.values[ind_next]}"
            )

        elif direction == "bw" and direction_next == "bw" and verbose:
            print(
                f"Missing forward measurement beween {ds_bw.time.values[ind]} and {ds_bw.time.values[ind_next]}"
            )

    # throw out is dt differs from its neighbors
    if verify_timedeltas:
        dt = (
            ds_bw.isel(time=iuse_chbw).time.values
            - ds_fw.isel(time=iuse_chfw).time.values
        ) / np.timedelta64(1, "s")
        leaveout = np.zeros_like(dt, dtype=bool)
        leaveout[1:-1] = np.isclose(dt[:-2], dt[2:], atol=1.5, rtol=0.0) * ~np.isclose(
            dt[:-2], dt[1:-1], atol=1.5, rtol=0.0
        )
        iuse_chfw2 = np.array(iuse_chfw)[~leaveout]
        iuse_chbw2 = np.array(iuse_chbw)[~leaveout]

        if verbose:
            for itfw, itbw in zip(
                np.array(iuse_chfw)[leaveout], np.array(iuse_chbw)[leaveout]
            ):
                print(
                    "The following measurements do not belong together, as the time difference\n"
                    "between the\forward and backward measurements is more than 1.5 seconds\n"
                    "larger than the neighboring measurements.\n"
                    f"FW: {ds_fw.isel(time=itfw).time.values} and BW: {ds_bw.isel(time=itbw).time.values}"
                )
        return ds_fw.isel(time=iuse_chfw2), ds_bw.isel(time=iuse_chbw2)

    return ds_fw.isel(time=iuse_chfw), ds_bw.isel(time=iuse_chbw)


def shift_double_ended(
    ds: xr.Dataset, i_shift: int, verbose: bool = True
) -> xr.Dataset:
    """The cable length was initially configured during the DTS measurement. For double ended
    measurements it is important to enter the correct length so that the forward channel and the
    backward channel are aligned.

    This function can be used to shift the backward channel so that the forward channel and the
    backward channel are aligned. The backward channel is shifted per index
    along the x dimension.

    The DataStore object that is returned only has values for the backscatter that are measured
    for both the forward channel and the backward channel in the shifted object

    There is no interpolation, as this would alter the accuracy.


    Parameters
    ----------
    ds : DataSore object
        DataStore object that needs to be shifted
    i_shift : int
        if i_shift < 0, the cable was configured to be too long and there is too much data
        recorded. If i_shift > 0, the cable was configured to be too short and part of the cable is
        not measured
    verbose: bool
        If True, the function will inform the user which variables are
        dropped. If False, the function will silently drop the variables.

    Returns:
    --------
    ds2 : DataStore object
        With a shifted x-axis
    """
    assert isinstance(i_shift, (int, np.integer))

    nx = ds.x.size
    nx2 = nx - i_shift
    if i_shift < 0:
        # The cable was configured to be too long.
        # There is too much data recorded.
        st = ds.st.data[:i_shift]
        ast = ds.ast.data[:i_shift]
        rst = ds.rst.data[-i_shift:]
        rast = ds.rast.data[-i_shift:]
        x2 = ds.x.data[:i_shift]
        # TMP2 = ds.tmp.data[:i_shift]

    else:
        # The cable was configured to be too short.
        # Part of the cable is not measured.
        st = ds["st"].data[i_shift:]
        ast = ds["ast"].data[i_shift:]
        rst = ds["rst"].data[:nx2]
        rast = ds["rast"].data[:nx2]
        x2 = ds["x"].data[i_shift:]
        # TMP2 = ds.tmp.data[i_shift:]

    d2_coords = dict(ds.coords)
    d2_coords["x"] = xr.DataArray(data=x2, dims=("x",), attrs=ds["x"].attrs)

    d2_data = dict(ds.data_vars)
    for k in ds.data_vars:
        if "x" in ds[k].dims and k in d2_data:
            del d2_data[k]

    new_data = (("st", st), ("ast", ast), ("rst", rst), ("rast", rast))

    for k, v in new_data:
        d2_data[k] = xr.DataArray(data=v, dims=ds[k].dims, attrs=ds[k].attrs)

    not_included = [k for k in ds.data_vars if k not in d2_data]
    if not_included and verbose:
        msg = (
            "The following variables could not be shifted and are not included in the "
            f"new dataset: {not_included}. You can silence this warning with the "
            "keyword argument `verbose=False`."
        )
        warnings.warn(msg, UserWarning)

    return xr.Dataset(data_vars=d2_data, coords=d2_coords, attrs=ds.attrs)


def suggest_cable_shift_double_ended(
    ds: xr.Dataset,
    irange: npt.NDArray[np.int_],
    plot_result: bool = True,
    **fig_kwargs,
) -> tuple[int, int]:
    """The cable length was initially configured during the DTS measurement.
    For double ended measurements it is important to enter the correct length
    so that the forward channel and the backward channel are aligned.

    This function can be used to find the shift of the backward channel so
    that the forward channel and the backward channel are aligned. The shift
    index refers to the x-dimension.

    The attenuation should be approximately a straight line with jumps at the
    splices. Temperature independent and invariant over time. The following
    objective functions seems to do the job at determining the best shift for
    which the attenuation is most straight.

    Err1 sums the first derivative. Is a proxy for the length of the
    attenuation line. Err2 sums the second derivative. Is a proxy for the
    wiggelyness of the line.

    The top plot shows the origional Stokes and the origional and shifted
    anti-Stokes The bottom plot is generated that shows the two objective
    functions


    Parameters
    ----------
    ds : Xarray Dataset
    irange : array-like
        a numpy array with data of type int. Containing all the shift index
        that are tested.
        Example: np.arange(-250, 200, 1, dtype=int). It shifts the return
        scattering with 250 indices. Calculates err1 and err2. Then shifts
        the return scattering with 249 indices. Calculates err1 and err2. The
        lowest err1 and err2 are suggested as best shift options.
    plot_result : bool
        Plot the summed error as a function of the shift.

    Returns:
    --------
    ishift1: int
        Suggested shift based on Err1
    ishift2: int
        Suggested shift based on Err2
    """
    err1 = []
    err2 = []

    for shift in irange:
        i_shift = int(shift)  # int() because iterating irange does not narrow type.
        nx = ds["x"].size
        nx2 = nx - i_shift
        if i_shift < 0:
            # The cable was configured to be too long. There is too much data recorded.
            st = ds["st"].data[:i_shift]
            ast = ds["ast"].data[:i_shift]
            rst = ds["rst"].data[-i_shift:]
            rast = ds["rast"].data[-i_shift:]
            x2 = ds["x"].data[:i_shift]
        else:
            # The cable was configured to be too short. Part of the cable is not measured.
            st = ds["st"].data[i_shift:]
            ast = ds["ast"].data[i_shift:]
            rst = ds["rst"].data[:nx2]
            rast = ds["rast"].data[:nx2]
            x2 = ds["x"].data[i_shift:]

        with warnings.catch_warnings():
            # Supress log(x/0) warnings. The data will result in NaNs values.
            warnings.filterwarnings(
                "ignore", message="invalid value encountered in log"
            )
            i_f = np.log(st / ast)
            i_b = np.log(rst / rast)

        att = (i_b - i_f) / 2  # varianble E in article

        att_dif1 = np.diff(att, n=1, axis=0)
        att_x_dif1 = 0.5 * x2[1:] + 0.5 * x2[:-1]
        err1_mask = np.logical_and(att_x_dif1 > 1.0, att_x_dif1 < 150.0)
        err1.append(np.nansum(np.abs(att_dif1[err1_mask])))

        att_dif2 = np.diff(att, n=2, axis=0)
        att_x_dif2 = x2[1:-1]
        err2_mask = np.logical_and(att_x_dif2 > 1.0, att_x_dif2 < 150.0)
        err2.append(np.nansum(np.abs(att_dif2[err2_mask])))

    # int() is required for typing.
    ishift1 = int(irange[np.argmin(err1, axis=0)])
    ishift2 = int(irange[np.argmin(err2, axis=0)])

    if plot_result:
        if fig_kwargs is None:
            fig_kwargs = {}

        f, (ax0, ax1) = plt.subplots(2, 1, sharex=False, **fig_kwargs)
        f.suptitle(f"best shift is {ishift1} or {ishift2}")

        dt = ds.isel(time=0)[["st", "ast", "rst", "rast", "x"]]
        x = dt["x"].data
        y = dt["st"].data
        ax0.plot(x, y, label="ST original")
        y = dt["rst"].data
        ax0.plot(x, y, label="REV-ST original")

        dtsh1 = shift_double_ended(dt, ishift1)
        dtsh2 = shift_double_ended(dt, ishift2)
        x1 = dtsh1["x"].data
        x2 = dtsh2["x"].data
        y1 = dtsh1["rst"].data
        y2 = dtsh2["rst"].data
        ax0.plot(x1, y1, label=f"ST i_shift={ishift1}")
        ax0.plot(x2, y2, label=f"ST i_shift={ishift2}")
        ax0.set_xlabel("x (m)")
        ax0.legend()

        ax2 = ax1.twinx()
        ax1.plot(irange, err1, c="red", label="1 deriv")
        ax2.plot(irange, err2, c="blue", label="2 deriv")
        ax1.axvline(
            ishift1, c="red", linewidth=0.8, label=f"1 deriv. i_shift={ishift1}"
        )
        ax2.axvline(
            ishift2, c="blue", linewidth=0.8, label=f"2 deriv. i_shift={ishift1}"
        )
        ax1.set_xlabel("i_shift")
        ax1.legend(loc=2)  # left axis
        ax2.legend(loc=1)  # right axis

        plt.tight_layout()

    return ishift1, ishift2


def ufunc_per_section_helper(
    sections=None,
    func=None,
    x_coords=None,
    reference_dataset=None,
    dataarray=None,
    subtract_from_dataarray=None,
    subtract_reference_from_dataarray=False,
    ref_temp_broadcasted=False,
    calc_per="stretch",
    **func_kwargs,
):
    """User function applied to parts of the cable. Super useful,
    many options and slightly
    complicated.

    The function `func` is taken over all the timesteps and calculated
    per `calc_per`. This
    is returned as a dictionary

    Parameters
    ----------
    sections : Dict[str, List[slice]], optional
        If `None` is supplied, `ds.sections` is used. Define calibration
        sections. Each section requires a reference temperature time series,
        such as the temperature measured by an external temperature sensor.
        They should already be part of the DataStore object. `sections`
        is defined with a dictionary with its keywords of the
        names of the reference temperature time series. Its values are
        lists of slice objects, where each slice object is a fiber stretch
        that has the reference temperature. Afterwards, `sections` is stored
        under `ds.sections`.
    func : callable, str
        A numpy function, or lambda function to apply to each 'calc_per'.
    x_coords : xarray.DataArray, optional
        x-coordinates, stored as ds.x. If supplied, returns the x-indices of
        the reference sections.
    reference_dataset : xarray.Dataset or Dict, optional
        Contains the reference temperature timeseries refered to in `sections`.
        Not required if `x_indices`.
    dataarray : xarray.DataArray, optional
        Pass your DataArray of which you want to compute the statistics. Has an
        (x,) dimension or (x, time) dimensions.
    subtract_from_dataarray : xarray.DataArray, optional
        Pass your DataArray of which you want to subtract from `dataarray` before you
        compute the statistics. Has an (x,) dimension or (x, time) dimensions.
    subtract_reference_from_dataarray : bool
        If True the reference temperature according to sections is subtracted from
        dataarray before computing statistics
    ref_temp_broadcasted : bool
        Use if you want to return the reference temperature of shape of the reference
        sections
    calc_per : {'all', 'section', 'stretch'}
    func_kwargs : dict
        Dictionary with options that are passed to func

    Returns:
    --------

    Examples:
    ---------
    1. Calculate the variance of the residuals in the along ALL the\
    reference sections wrt the temperature of the water baths

    >>> tmpf_var = ufunc_per_section_helper(
    >>>     func='var',
    >>>     calc_per='all',
    >>>     dataarray=d['tmpf'],
    >>>     subtract_reference_from_dataarray=True)

    2. Calculate the variance of the residuals in the along PER\
    reference section wrt the temperature of the water baths

    >>> tmpf_var = ufunc_per_section_helper
    >>>     func='var',
    >>>     calc_per='stretch',
    >>>     dataarray=d['tmpf'],
    >>>     subtract_reference_from_dataarray=True)

    3. Calculate the variance of the residuals in the along PER\
    water bath wrt the temperature of the water baths

    >>> tmpf_var = ufunc_per_section_helper(
    >>>     func='var',
    >>>     calc_per='section',
    >>>     dataarray=d['tmpf'],
    >>>     subtract_reference_from_dataarray=True)

    4. Obtain the coordinates of the measurements per section

    >>> locs = ufunc_per_section_helper(
    >>>     func=None,
    >>>     dataarray=d.x,
    >>>     subtract_reference_from_dataarray=False,
    >>>     ref_temp_broadcasted=False,
    >>>     calc_per='stretch')

    5. Number of observations per stretch

    >>> nlocs = ufunc_per_section_helper(
    >>>     func=len,
    >>>     dataarray=d.x,
    >>>     subtract_reference_from_dataarray=False,
    >>>     ref_temp_broadcasted=False,
    >>>     calc_per='stretch')

    6. broadcast the temperature of the reference sections to\
    stretch/section/all dimensions. The value of the reference\
    temperature (a timeseries) is broadcasted to the shape of self[\
    label]. The dataarray is not used for anything else.

    >>> temp_ref = ufunc_per_section_helper(
    >>>     dataarray=d["st"],
    >>>     ref_temp_broadcasted=True,
    >>>     calc_per='all')

    7. x-coordinate index

    >>> ix_loc = ufunc_per_section_helper(x_coords=d.x)


    Note:
    ----
    If `dataarray` or `subtract_from_dataarray` is a Dask array, a Dask
    array is returned else a numpy array is returned
    """
    if not func:

        def func(a):
            """Parameters
            ----------
            a

            Returns:
            -------

            """
            return a

    elif isinstance(func, str) and func == "var":

        def func(a):
            """Parameters
            ----------
            a

            Returns:
            -------

            """
            return np.var(a, ddof=1)

    else:
        assert callable(func)

    assert calc_per in ["all", "section", "stretch"]
    assert "x_indices" not in func_kwargs, "pass x_coords arg instead"

    if x_coords is None and (
        (dataarray is not None and hasattr(dataarray.data, "chunks"))
        or (subtract_from_dataarray and hasattr(subtract_from_dataarray.data, "chunks"))
    ):
        concat = da.concatenate
    else:
        concat = np.concatenate

    out = dict()

    for k, section in sections.items():
        out[k] = []
        for stretch in section:
            if x_coords is not None:
                # get indices from stretches
                assert subtract_from_dataarray is None
                assert not subtract_reference_from_dataarray
                assert not ref_temp_broadcasted
                assert not func_kwargs, "Unsupported kwargs"

                # so it is slicable with x-indices
                _x_indices = x_coords.astype(int) * 0 + np.arange(x_coords.size)
                arg1 = _x_indices.sel(x=stretch).data
                out[k].append(arg1)

            elif (
                subtract_from_dataarray is not None
                and not subtract_reference_from_dataarray
                and not ref_temp_broadcasted
            ):
                # calculate std wrt other series
                arg1 = dataarray.sel(x=stretch).data
                arg2 = subtract_from_dataarray.sel(x=stretch).data
                out[k].append(arg1 - arg2)

            elif (
                subtract_from_dataarray is None
                and subtract_reference_from_dataarray
                and not ref_temp_broadcasted
            ):
                # calculate std wrt reference temperature of the corresponding bath
                arg1 = dataarray.sel(x=stretch).data
                arg2 = reference_dataset[k].data
                out[k].append(arg1 - arg2)

            elif (
                subtract_from_dataarray is None
                and not subtract_reference_from_dataarray
                and ref_temp_broadcasted
            ):
                # Broadcast the reference temperature to the length of the stretch
                arg1 = dataarray.sel(x=stretch).data
                arg2 = da.broadcast_to(reference_dataset[k].data, arg1.shape)
                out[k].append(arg2)

            elif (
                subtract_from_dataarray is None
                and not subtract_reference_from_dataarray
                and not ref_temp_broadcasted
            ):
                # calculate std wrt mean value
                arg1 = dataarray.sel(x=stretch).data
                out[k].append(arg1)

        if calc_per == "stretch":
            out[k] = [func(argi, **func_kwargs) for argi in out[k]]

        elif calc_per == "section":
            # flatten the out_dict to sort them
            start = [i.start for i in section]
            i_sorted = np.argsort(start)
            out_flat_sort = [out[k][i] for i in i_sorted]
            out[k] = func(concat(out_flat_sort), **func_kwargs)

        elif calc_per == "all":
            pass

    if calc_per == "all":
        # flatten the out_dict to sort them
        start = [item.start for sublist in sections.values() for item in sublist]
        i_sorted = np.argsort(start)
        out_flat = [item for sublist in out.values() for item in sublist]
        out_flat_sort = [out_flat[i] for i in i_sorted]
        out = func(concat(out_flat_sort, axis=0), **func_kwargs)

        if hasattr(out, "chunks") and len(out.chunks) > 0 and "x" in dataarray.dims:
            # also sum the chunksize in the x dimension
            # first find out where the x dim is
            ixdim = dataarray.dims.index("x")
            c_old = out.chunks
            c_new = list(c_old)
            c_new[ixdim] = sum(c_old[ixdim])
            out = out.rechunk(c_new)

    return out
